import attr
from lingpy import Wordlist
import pyclts

import pytest

from pylexibank import lingpy_util
from pylexibank import models as pbds


def test_iter_cognates_and_alignments(dataset_cldf):
    assert not list(lingpy_util.iter_cognates(
        dataset_cldf.cldf_specs().get_writer(dataset=dataset_cldf), method='sca'))

    ds = dataset_cldf.cldf_specs().get_dataset()
    res = list(lingpy_util.iter_cognates(ds, method='lexstat'))
    assert res
    lingpy_util.iter_alignments(ds, res)
    assert 'Alignment' in res[0]

    ds = dataset_cldf.cldf_specs().get_dataset()
    res = list(lingpy_util.iter_cognates(ds, method='lexstat'))
    lingpy_util.iter_alignments(
        lingpy_util._cldf2wordlist(ds),
        res,
        almkw=dict(
            ref='lid',
            row='parameter_id',
            transcription='form',
            segments='segments',
            col='language_id'))
    assert 'Alignment' in res[0]


def test_wordlist2cognates(repos, mocker, dataset):
    @attr.s
    class Lexeme(pbds.Lexeme):
        Concept = attr.ib(default=None)
        Segments = attr.ib(default=[])
    @attr.s
    class Lexeme2(pbds.Lexeme):
        Concept = attr.ib(default=None)

    dsdir = repos / 'datasets' / 'test_dataset'
    if not dsdir.joinpath('cldf').exists():
        dsdir.joinpath('cldf').mkdir()
    dataset.cognate_class = pbds.Cognate
    dataset.language_class = pbds.Language
    dataset.concept_class = pbds.Concept
    dataset.split_forms = lambda _, s: [s]
    dataset.dir = dsdir
    dataset.tr_analyses = {}
    dataset.cldf_dir = dsdir.joinpath('cldf')

    dataset.lexeme_class = Lexeme
    with dataset.cldf_writer(mocker.Mock()) as ds:
        # needs to be fixed XXX
        ds.tokenize = lambda _, x: []
        ds.add_forms_from_value(
            Value='form,form2',
            Concept='meaning',
            Language_ID='1',
            Parameter_ID='p'
        )

    dataset.lexeme_class = Lexeme2
    with dataset.cldf_writer(mocker.Mock(clts=mocker.Mock(api=pyclts.CLTS(repos)))) as ds2:
        ds2.add_form_with_segments(
            Value='form,form2',
            Concept='meaning',
            Language_ID='1',
            Parameter_ID='p',
            Form='form',
            Segments=['f', 'o']
        )
        # needs to be fixed XXX
        ds2.tokenize = lambda _, x: [x]
        ds2.add_form(
            Value='form,form2',
            Concept='meaning',
            Language_ID='1',
            Parameter_ID='p',
            Form='form',
        )
    # lid, ipa, concept
    wl = Wordlist(lingpy_util._cldf2wld(ds2), row='concept', col='language_id')
    res = list(lingpy_util.wordlist2cognates(wl, 'src'))
    assert isinstance(res[0], dict)



"""
Check lexibank plumbing for lexibank datasets
"""
import functools
import collections

from cldfbench.cli_util import with_datasets

from pylexibank.cli_util import add_dataset_spec, warning


def register(parser):
    add_dataset_spec(parser, multiple=True)


def check(ds, args, warnings=None):
    warn = functools.partial(warning, args, dataset=ds, warnings=warnings)

    args.log.info('checking {0} - plumbing'.format(ds))

    # check that there's a description of contributions:
    if not ds.get_creators_and_contributors(strict=True)[0]:
        warn('Dataset does not describe creators in {}'.format(ds.contributors_path))

    # check that there's no local concepts.csv:
    if (ds.dir / 'etc' / 'concepts.csv').exists():  # pragma: no cover
        warn('Dataset uses a local ./etc/concepts.csv rather than a conceptlist from Concepticon')

    cldf = ds.cldf_reader()
    # check lexemes.csv
    if (ds.dir / 'etc' / 'lexemes.csv').exists():
        try:
            values = set(f['Value'] for f in cldf['FormTable'])
            for r in ds.lexemes:
                if ds.lexemes[r] and r not in values:  # replacement of form x -> y
                    warn("lexemes.csv contains un-needed conversion '{0}' -> '{1}'".format(
                        r, ds.lexemes[r]))
                if not ds.lexemes[r] and r in values:  # removal of form x -> ""
                    warn("lexemes.csv contains un-handled removal '{0}' -> '{1}'".format(
                        r, ds.lexemes[r]))
        except KeyError:  # pragma: no cover
            warn('Dataset does not seem to be a lexibank dataset - FormTable has no Value column!')

    if (not getattr(ds, 'cross_concept_cognates', False)) and cldf.get('CognateTable'):
        # check that there are no cross-concept cognate sets:
        id_col = cldf['FormTable', 'id'].name
        pid_col = cldf['FormTable', 'parameterReference'].name
        fid_col = cldf['CognateTable', 'formReference'].name
        cogid_col = cldf['CognateTable', 'cognatesetReference'].name

        form_to_concept = {form[id_col]: form[pid_col] for form in cldf['FormTable']}
        cogset_to_concepts = collections.defaultdict(set)
        for cog in cldf['CognateTable']:
            cid = form_to_concept[cog[fid_col]]
            cogid = cog[cogid_col]
            if len(cogset_to_concepts[cogid]) == 1 \
                    and cid not in cogset_to_concepts[cogid]:
                # We warn when a second concept ID is detected for a cognateset.
                warn('Cross-concept cognate set {0}'.format(cogid))
            cogset_to_concepts[cogid].add(cid)


def run(args):
    with_datasets(args, check)



import pytest

from pylexibank import Dataset
from pylexibank.report import build_status_badge, report


@pytest.fixture
def ds(tmp_path, git_repo_factory):
    git_repo_factory(tmp_path)
    ds_ = Dataset()
    ds_.dir = tmp_path
    ds_.dir.joinpath('.travis.yml').write_text('#', encoding='utf8')
    ds_.dir.joinpath('NOTES.md').write_text('#', encoding='utf8')
    return ds_


def test_no_travis_badge():
    assert not build_status_badge(Dataset())


def test_travis_badge(ds):
    assert build_status_badge(ds)


def test_report(ds):
    assert report(ds)


def test_report_full(dataset):
    assert report(dataset)



"""
Create an initial orthography profile, seeded from the forms created by a first run
of lexibank.makecldf.
"""
from lingpy import Wordlist
from lingpy.sequence import profile
from cldfbench.cli_util import get_dataset, add_catalog_spec
from csvw.dsv import UnicodeWriter
from clldutils.clilib import ParserError

from pylexibank.cli_util import add_dataset_spec


def register(parser):
    add_dataset_spec(parser)
    add_catalog_spec(parser, 'clts')
    parser.add_argument(
        '--context',
        action='store_true',
        help='Create orthography profile with context',
        default=False)
    parser.add_argument(
        '-f', '--force',
        action='store_true',
        help='Overwrite existing profile',
        default=False)
    parser.add_argument(
        '--semi-diacritics',
        default='hsʃ̢ɕʂʐʑʒw',
        help='Indicate characters which can occur both as "diacritics" (second part in a sound) '
             'or alone.')
    parser.add_argument(
        '--merge-vowels',
        action='store_true',
        help='Indicate whether consecutive vowels should be merged.',
        default=False)
    parser.add_argument(
        '--dont-merge-geminates',
        action='store_true',
        default=False)


def run(args):
    bipa = args.clts.api.bipa
    func = profile.simple_profile
    cols = ['Grapheme', 'IPA', 'Frequence', 'Codepoints']
    kw = {
        'ref': 'form',
        'clts': bipa,
        'semi_diacritics': args.semi_diacritics,
        'merge_vowels': args.merge_vowels,
        'merge_geminates': not args.dont_merge_geminates,
    }
    if args.context:
        func = profile.context_profile
        cols = ['Grapheme', 'IPA', 'Examples', 'Languages', 'Frequence', 'Codepoints']
        kw['col'] = 'language_id'

    ds = get_dataset(args)
    profile_path = ds.etc_dir / 'orthography.tsv'
    if profile_path.exists() and not args.force:
        raise ParserError('Orthography profile exists already. To overwrite, pass "-f" flag')

    header, D = [], {}
    for i, row in enumerate(ds.cldf_reader()['FormTable'], start=1):
        if i == 1:
            header = [f for f in row.keys() if f != 'ID']
            D = {0: ['lid'] + [h.lower() for h in header]}

        row['Segments'] = ' '.join(row['Segments'])
        D[i] = [row['ID']] + [row[h] for h in header]

    with UnicodeWriter(profile_path, delimiter='\t') as writer:
        writer.writerow(cols)
        for row in func(Wordlist(D, row='parameter_id', col='language_id'), **kw):
            writer.writerow(row)
    args.log.info('Orthography profile written to {0}'.format(profile_path))



import shlex
import logging
import argparse

import pytest

from csvw import dsv
from clldutils import jsonlib
from cldfbench.__main__ import main
from pylexibank import cli_util


def test_warning(caplog):
    cli_util.warning(
        argparse.Namespace(log=logging.getLogger(__name__)),
        'message in a bottle',
        dataset=argparse.Namespace(id='abc')
    )
    assert caplog.records[-1].levelname == 'WARNING'


def _main(cmd, **kw):
    kw.setdefault('log', logging.getLogger(__name__))
    main(['--no-config'] + shlex.split(cmd), **kw)


@pytest.mark.filterwarnings("ignore:distutils Version:DeprecationWarning")
def test_makecldf_concepticon_concepts(repos):
    d = repos / 'datasets' / 'test_dataset_concepticon_concepts'
    _main('lexibank.makecldf {0} --glottolog {1} --concepticon {1} --clts {1}'.format(
        d / 'tdcc.py', repos))
    assert d.joinpath('cldf', 'parameters.csv').read_text(encoding='utf8').splitlines()[0] == \
        'ID,Name,Concepticon_ID,Concepticon_Gloss,NUMBER,ENGLISH,CHINESE,PAGE'


def test_makecldf_multi_profiles(repos):
    d = repos / 'datasets' / 'test_dataset_multi_profile'
    _main('lexibank.makecldf {0} --glottolog {1} --concepticon {1} --clts {1}'.format(
        str(d / 'tdmp.py'),
        str(repos),
    ))
    assert len(list(dsv.reader(d / 'cldf' / 'languages.csv', dicts=True))) == 1, \
        'languages not pruned!'
    assert len(list(dsv.reader(d / 'cldf' / 'parameters.csv', dicts=True))) == 1, \
        'parameters not pruned!'

    forms = list(dsv.reader(d / 'cldf' / 'forms.csv', dicts=True))
    assert forms[0]['Profile'] == 'p1'

    _main('lexibank.format_profile {0} --clts {1} --sort --trim --augment'.format(
        str(d / 'tdmp.py'),
        str(repos),
    ))
    assert 'FREQUENCY' in (d / 'etc' / 'orthography' / 'p1.tsv').read_text(encoding='utf8')


def test_makecldf(repos, dataset, dataset_cldf, dataset_no_cognates, sndcmp, capsys, tmp_path):
    _main('lexibank.makecldf {0} --glottolog {1} --concepticon {1} --clts {1}'.format(
        str(dataset.dir / 'td.py'),
        str(repos),
    ))
    assert 'Papunesia' in dataset.cldf_dir.joinpath('languages.csv').read_text(encoding='utf8')
    # Metadata for Zenodo is merged if this makes sense:
    assert len(jsonlib.load(dataset.dir / '.zenodo.json')['communities']) == 3

    _main('lexibank.makecldf {0} --dev --glottolog {1} --concepticon {1} --clts {1}'.format(
        str(dataset.dir / 'td.py'),
        str(repos),
    ))
    assert 'Papunesia' not in dataset.cldf_dir.joinpath('languages.csv').read_text(encoding='utf8')
    assert '### Replacement' in dataset.dir.joinpath('FORMS.md').read_text(encoding='utf8')

    _main('lexibank.makecldf {0} --glottolog {1} --concepticon {1} --clts {1}'.format(
        str(sndcmp.dir / 'ts.py'),
        str(repos),
    ))
    assert 'Bislama_Gloss' in sndcmp.cldf_dir.joinpath('parameters.csv').read_text(encoding='utf8')
    assert 'e56a5fc78ae5a66e783c17bc30019568' in sndcmp.cldf_dir.joinpath('media.csv').read_text(encoding='utf8')

    _main('lexibank.makecldf {0} --glottolog {1} --concepticon {1} --clts {1}'.format(
        str(dataset_cldf.dir / 'tdc.py'),
        str(repos),
    ))
    capout = capsys.readouterr().out
    assert 'The dataset has no sources' not in capout

    _main('lexibank.makecldf {0} --glottolog {1} --concepticon {1} --clts {1}'.format(
        str(dataset_no_cognates.dir / 'tdn.py'),
        str(repos),
    ))
    assert not dataset_no_cognates.cldf_dir.joinpath('cognates.csv').exists()
    _main('lexibank.load --db {3} {0} --glottolog {1} --concepticon {2}'.format(
        str(dataset_no_cognates.dir / 'tdn.py'),
        str(repos),
        str(repos),
        str(tmp_path / 'db'),
    ))


def test_check(dataset_cldf, caplog):
    _main('lexibank.check {0}'.format(str(dataset_cldf.dir / 'tdc.py')),
          log=logging.getLogger(__name__))
    assert [r for r in caplog.records if 'CONTRIBUTORS.md' in r.message]


def test_check_lexibank(dataset_cldf, caplog):
    _main(
        'lexibank.check_lexibank {0}'.format(str(dataset_cldf.dir / 'tdc.py')),
        log=logging.getLogger(__name__))
    warnings = [r.message for r in caplog.records if r.levelname == 'WARNING']
    print(warnings)
    assert any('Cross-concept' in w for w in warnings)


def test_ls(repos, tmp_path, dataset):
    _main('lexibank.load --db {3} {0} --glottolog {1} --concepticon {2}'.format(
        dataset.dir / 'td.py', repos, repos, tmp_path / 'db'))
    _main('lexibank.ls {0} --all --db {1}'.format(dataset.dir / 'td.py', tmp_path / 'db'))
    _main('lexibank.unload --db {1} {0}'.format(dataset.dir / 'td.py', tmp_path / 'db'))


def test_db(tmp_path, mocker):
    mocker.patch('pylexibank.commands.db.subprocess', mocker.Mock(return_value=0))
    _main('lexibank.db --db {0}'.format(tmp_path / 'db'))


def test_check_phonotactics(dataset, capsys):
    _main('lexibank.check_phonotactics {0}'.format(str(dataset.dir / 'td.py')))
    out, _ = capsys.readouterr()
    assert out.strip() == """| Type | ID | Value | Form | Graphemes | Segments |
|-------:|:---------------|:--------|:-------|:------------|:------------|
| 1 | lang1-param1-1 | a b; c | a b | | a _ b |
| 1 | lang2-param2-2 | a~b-c | ab | | + a + + b + |"""


def test_consonant_clusters(dataset, repos, caplog, capsys):
    d = repos / 'datasets' / 'test_dataset_multi_profile_with_cldf'
    _main('lexibank.consonant_clusters -l 4 {0} --clts {1}'.format(str(d / 'tdmpcldf.py'), repos))
    warnings = [r.message for r in caplog.records if r.levelname == 'WARNING']
    out, _ = capsys.readouterr()

    assert any("Found 1 potentially problematic consonant cluster(s) with length 4" in w for w in warnings)
    assert out.strip() == """| Language_ID | Length | Cluster | Words |
|:--------------|---------:|:----------------------|:-------------------------------|
| lang1 | 6 | ɡ̤ː ɡ̤ː b dʱʷ dʱʷ dʱʷ | ɡ̤ː ɡ̤ː b dʱʷ dʱʷ dʱʷ // axdou |"""


def test_check_profile(dataset, repos, caplog, capsys):
    _main('lexibank.check_profile {0} --clts {1}'.format(str(dataset.dir / 'td.py'), repos))
    assert len(caplog.records) == 2
    d = repos / 'datasets' / 'test_dataset_multi_profile_with_cldf'
    _main('lexibank.check_profile {0} --clts {1}'.format(str(d / 'tdmpcldf.py'), repos))
    assert len(caplog.records) == 4
    out, _ = capsys.readouterr()
    assert all(w in out for w in 'modified generated slashed unknown missing'.split())


def test_init_profile(dataset, repos):
    _main('lexibank.init_profile {0} --clts {1} -f --context --merge-vowels'.format(
        str(dataset.dir / 'td.py'), repos))
    with pytest.raises(SystemExit):
        _main('lexibank.init_profile {0} --clts {1}'.format(str(dataset.dir / 'td.py'), repos))


def test_language_profiles(dataset_cldf, repos):
    with pytest.raises(ValueError):
        _main('lexibank.language_profiles {0}'.format(str(dataset_cldf.dir / 'tdc.py')))
    d = repos / 'datasets' / 'test_dataset_cldf_capitalisation'
    _main('lexibank.language_profiles {0}'.format(str(d / 'tdc.py')))


def test_readme(dataset, repos):
    _main(
        'lexibank.readme {0} --glottolog {1}'.format(str(dataset.dir / 'td.py'), repos))
    assert dataset.dir.joinpath('FORMS.md').exists()
    assert '# Contributors' in dataset.dir.joinpath('README.md').read_text(encoding='utf8')


def test_new(tmp_path, mocker):
    mocker.patch('cldfbench.metadata.input', mocker.Mock(return_value='abc'))
    _main('new --template lexibank_simple --out ' + str(tmp_path))
    assert tmp_path.joinpath('abc', 'CONTRIBUTORS.md').exists()

    mocker.patch('cldfbench.metadata.input', mocker.Mock(return_value='cde'))
    _main('new --template lexibank_combined --out ' + str(tmp_path))
    assert '{{' not in tmp_path.joinpath('cde', 'lexibank_cde.py').read_text(encoding='utf8')




import io
import shutil
import zipfile

import pytest

import attr
from csvw.dsv import reader
from pylexibank.providers.sndcmp import SNDCMP
from pylexibank.providers.sndcmp import SNDCMPConcept


JSONSTR = '{"languages":[{"FilePathPart": "Br_Tup_MaAwTG_Aweti_Aweti_Saitao_Dl"}]}'
JSONZIP = '{"EAEA0-4DE0-B3E9-31CE-0":{"metadata":'\
          '{"name":"Br_Tup_MaAwTG_Aweti_Aweti_Saitao_Dl_001_um_one"}}}'


@pytest.fixture
def sndcmp_dir(tmp_path, repos):
    shutil.copytree(repos / 'datasets' / 'sndcmp', tmp_path / 'sndcmp')
    return tmp_path / 'sndcmp'


@pytest.fixture
def sndcmp_dataset(repos, glottolog, concepticon, sndcmp_dir):
    class CustomConcept(SNDCMPConcept):
        Bislama_Gloss = attr.ib(default=None)

    class Dataset(SNDCMP):
        dir = sndcmp_dir
        id = "sndcmpvanuatu"
        study_name = "Vanuatu"
        second_gloss_lang = "Bislama"
        source_id_array = ["Shimelman2019"]
        create_cognates = True
        concept_class = CustomConcept
        form_placeholder = 'P'
        only_proto_forms = True

        def get_source_id_array(self, lexeme):
            return ['ab']

    return Dataset()


@pytest.fixture
def sndcmp2_dataset(repos, glottolog, concepticon, sndcmp_dir):
    class Dataset(SNDCMP):
        dir = sndcmp_dir
        id = "sndcmpvanuatu"
        study_name = "Vanuatu"
        source_id_array = ["Shimelman2019"]
        create_cognates = False

    return Dataset()


@pytest.fixture
def sndcmp_dl_dataset(repos, glottolog, concepticon, sndcmp_dir):
    class Dataset(SNDCMP):
        dir = sndcmp_dir
        id = "sndcmpbrazil"
        study_name = "Brazil"
        second_gloss_lang = None
        source_id_array = ["xy"]
        create_cognates = False

    return Dataset()


def test_sndcmp(sndcmp_dataset, mocker):
    sndcmp_dataset.cmd_create_ref_etc_files(mocker.MagicMock())
    assert (sndcmp_dataset.raw_dir / 'languages.csv').exists()
    assert (sndcmp_dataset.raw_dir / 'concepts.csv').exists()
    csv = sndcmp_dataset.raw_dir / 'concepts.csv'
    res = list(reader(csv, dicts=True))
    assert len(res) == 3
    assert 'Bislama_Gloss' in res[0]
    assert res[0]["IndexInSource"] == '1-0'


def test_sndcmp2(sndcmp2_dataset, mocker):
    sndcmp2_dataset.cmd_create_ref_etc_files(mocker.MagicMock())
    csv = sndcmp2_dataset.raw_dir / 'concepts.csv'
    res = list(reader(csv, dicts=True))
    assert len(res) == 3
    assert 'Bislama_Gloss' not in res[0]


def test_sndcmp_cldf(sndcmp_dataset, mocker):
    sndcmp_dataset.cmd_makecldf(mocker.MagicMock())


def test_sndcmp_dl(sndcmp_dl_dataset, mocker):
    class Requests(mocker.Mock):
        def get(self, *args, **kw):
            if 'zip' in args[0]:
                s = io.BytesIO()
                z = zipfile.ZipFile(s, 'w')
                z.writestr('catalog.json', JSONZIP.encode('utf8'))
                z.close()
                return mocker.Mock(
                    status_code=200,
                    iter_content=mocker.Mock(return_value=[s.getvalue()]))
            else:
                return mocker.Mock(
                    status_code=200,
                    iter_content=mocker.Mock(return_value=[JSONSTR.encode('utf8')]))

    mocker.patch('cldfbench.datadir.requests', Requests())

    sndcmp_dl_dataset.cmd_download(mocker.Mock())
    assert (sndcmp_dl_dataset.raw_dir / 'brazil.json').exists()



import re
import logging
import pathlib
import textwrap
import itertools
import collections
import collections.abc

import attr
from csvw.metadata import Column
from pycldf.dataset import Wordlist
import pyclts.models

from cldfbench.cldf import CLDFWriter
from cldfbench.util import iter_requirements, get_entrypoints

from pylexibank.transcription import Analysis, analyze, valid_sequence
from pylexibank.util import iter_repl, get_concepts, get_ids_and_attrs

__all__ = ['LexibankWriter']
log = logging.getLogger('pylexibank')

MD_NAME = 'cldf-metadata.json'
ID_PATTERN = re.compile(r'[A-Za-z0-9_\-]+$')


@attr.s
class Options:
    keep_languages = attr.ib(default=False, validator=attr.validators.instance_of(bool))
    keep_parameters = attr.ib(default=False, validator=attr.validators.instance_of(bool))


class LexibankWriter(CLDFWriter):
    def __init__(self, dataset=None, **kw):
        super().__init__(dataset=dataset, **kw)
        self._count = collections.defaultdict(int)
        self._cognate_count = collections.defaultdict(int)
        self.options = Options(**getattr(dataset, 'writer_options', {}))

    def write(self, **kw):
        from pylexibank import ENTRY_POINT

        self.cldf.add_provenance(wasGeneratedBy=[
            collections.OrderedDict([
                ('dc:title', "lingpy-rcParams"), ('dc:relation', 'lingpy-rcParams.json')])])

        super().write(**kw)
        # We rewrite requirements.txt, excluding all lexibank dataset modules:
        exclude = {'egg=' + ep.load().__module__ for ep in get_entrypoints(ENTRY_POINT)}
        reqs = []
        for req in iter_requirements():
            if not any(mod in req for mod in exclude):
                reqs.append(req)

        self.cldf_spec.dir.joinpath('requirements.txt').write_text('\n'.join(reqs), encoding='utf8')

    def __enter__(self):
        super().__enter__()
        default_cldf = Wordlist.from_metadata(pathlib.Path(__file__).parent / MD_NAME)

        self._obj_index = {}
        for cls in [
            self.dataset.lexeme_class,
            self.dataset.language_class,
            self.dataset.concept_class,
            self.dataset.cognate_class,
        ]:
            if cls.__doc__:
                self.cldf[cls.__cldf_table__()].common_props['dc:description'] = \
                    textwrap.dedent(cls.__doc__)

            self.objects[cls.__cldf_table__()] = []
            self._obj_index[cls.__cldf_table__()] = set()

            cols = set(
                col.header for col in self.cldf[cls.__cldf_table__()].tableSchema.columns)
            properties = set(
                col.propertyUrl.uri for col in self.cldf[cls.__cldf_table__()].tableSchema.columns
                if col.propertyUrl)
            for field in cls.fieldnames():
                try:
                    col = default_cldf[cls.__cldf_table__(), field]
                    #
                    # We added Latitude and Longitude to the default metadata later, and want to
                    # make sure, existing datasets are upgraded silently.
                    #
                    if field in ['Latitude', 'Longitude'] \
                            and cls.__cldf_table__() == 'LanguageTable':  # pragma: no cover
                        properties.add(col.propertyUrl.uri)
                        self.cldf[cls.__cldf_table__(), field].propertyUrl = col.propertyUrl
                        self.cldf[cls.__cldf_table__(), field].datatype = col.datatype
                except KeyError:
                    kw = {k: v for k, v in attr.fields_dict(cls)[field].metadata.items()}
                    kw.setdefault('datatype', 'string')
                    kw.setdefault('name', field)
                    col = Column.fromvalue(kw)
                if (col.propertyUrl and col.propertyUrl.uri not in properties) or \
                        ((not col.propertyUrl) and (field not in cols)):
                    self.cldf[cls.__cldf_table__()].tableSchema.columns.append(col)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for table in ['FormTable', 'LanguageTable', 'ParameterTable']:
            self.objects.setdefault(table, [])
        if not self.objects.get('CognateTable'):
            self.cldf.tablegroup.tables = [
                t for t in self.cldf.tables if str(t.url) != 'cognates.csv']
            if 'CognateTable' in self.objects:
                del self.objects['CognateTable']
        for fk, table in [('Parameter_ID', 'ParameterTable'), ('Language_ID', 'LanguageTable')]:
            if (table == 'ParameterTable' and self.options.keep_parameters) or \
                    (table == 'LanguageTable' and self.options.keep_languages):
                continue  #
            # If opt-in, we only add concepts and languages that are referenced by forms.
            refs = set(obj[fk] for obj in self.objects['FormTable'])
            self.objects[table] = [obj for obj in self.objects[table] if obj['ID'] in refs]
        super().__exit__(exc_type, exc_val, exc_tb)

    def add_sources(self, *args):
        if not args and self.dataset.raw_dir.joinpath('sources.bib').exists():
            args = self.dataset.raw_dir.read_bib()
        self.cldf.sources.add(*args)

    def lexeme_id(self, kw):
        self._count[(kw['Language_ID'], kw['Parameter_ID'])] += 1
        return '{0}-{1}-{2}'.format(
            kw['Language_ID'],
            kw['Parameter_ID'],
            self._count[(kw['Language_ID'], kw['Parameter_ID'])])

    def cognate_id(self, kw):
        self._cognate_count[kw['Form_ID']] += 1
        return '{0}-{1}'.format(kw['Form_ID'], self._cognate_count[kw['Form_ID']])

    def tokenize(self, item, string, **kw):
        if self.dataset.tokenizer:
            return self.dataset.tokenizer(item, string, **kw)

    def add_form_with_segments(self, **kw):
        """
        :return: dict with the newly created lexeme
        """
        # Do we have morpheme segmentation on top of phonemes?
        with_morphemes = '+' in self['FormTable', 'Segments'].separator

        language, concept, value, form, segments = (kw.get(v) for v in [
            'Language_ID', 'Parameter_ID', 'Value', 'Form', 'Segments'])

        # check for required kws
        if not all([language, concept, value, form, segments]):
            raise ValueError('language, concept, value, form, and segments must be supplied')

        # Correct segments according to mapping in etc/segments.csv:
        for k, v in self.dataset.segments.items():
            segments = list(iter_repl(segments, k.split(), v.split()))
        kw['Segments'] = segments
        kw.update(ID=self.lexeme_id(kw), Form=form)
        lexeme = self._add_object(self.dataset.lexeme_class, **kw)

        analysis = self.dataset.tr_analyses.setdefault(kw['Language_ID'], Analysis())
        try:
            segments = kw['Segments']
            if with_morphemes:
                segments = list(itertools.chain(*[s.split() for s in segments]))
            valid = valid_sequence(segments)
            _, _bipa, _sc, _analysis = analyze(self.args.clts.api, segments, analysis)

            # update the list of `bad_words` if necessary; we precompute a
            # list of data types in `_bipa` just to make the conditional
            # checking easier
            _bipa_types = [type(s) for s in _bipa]
            if (pyclts.models.UnknownSound in _bipa_types) or '?' in _sc or not valid:
                self.dataset.tr_bad_words.append(kw)
        except ValueError:  # pragma: no cover
            self.dataset.tr_invalid_words.append(kw)
        except (KeyError, AttributeError):  # pragma: no cover
            print(kw['Form'], kw)
            raise
        return lexeme

    def add_form(self, with_morphemes=False, **kw):
        """
        :return: dict with the newly created form
        """
        language, concept, value, form = [kw.get(v) for v in [
            'Language_ID', 'Parameter_ID', 'Value', 'Form']]

        # check for required kws
        if language is None or concept is None or value is None \
                or form is None:
            raise ValueError('language, concept, value, and form must be supplied')

        # check for kws not allowed
        if 'Segments' in kw:
            raise ValueError('segmented data must be passed with add_form_with_segments')

        # point to difference in value and form
        if form != value:
            log.debug('iter_forms split: "{0}" -> "{1}"'.format(value, form))

        if form and form not in self.dataset.form_spec.missing_data:
            # try to segment the data now
            profile = kw.pop('profile', None)
            kw.setdefault(
                'Segments',
                self.tokenize(kw, form, **(dict(profile=profile) if profile else {})) or [])
            if kw['Segments']:
                return self.add_form_with_segments(**kw)

            kw.update(ID=self.lexeme_id(kw), Form=form)
            return self._add_object(self.dataset.lexeme_class, **kw)

    def add_forms_from_value(self, split_value=None, **kw):
        """
        :return: list of dicts corresponding to newly created Lexemes
        """
        lexemes = []

        if 'Segments' in kw:
            raise ValueError('segmented data must be passed with add_form_with_segments')
        if 'Form' in kw:
            raise ValueError('forms must be passed with add_form')

        svkw = {}
        if split_value is None:
            split_value = self.dataset.form_spec.split
            svkw = dict(lexemes=self.dataset.lexemes)

        # Do we have morpheme segmentation on top of phonemes?
        with_morphemes = '+' in self['FormTable', 'Segments'].separator

        for i, form in enumerate(split_value(kw, kw['Value'], **svkw)):
            kw_ = kw.copy()
            kw_['Form'] = form
            if kw_['Form']:
                kw_ = self.add_form(with_morphemes=with_morphemes, **kw_)
                if kw_:
                    lexemes.append(kw_)

        return lexemes

    def add_lexemes(self, split_value=None, **kw):
        """
        :return: list of dicts corresponding to newly created Lexemes
        """
        return self.add_forms_from_value(split_value=split_value, **kw)

    def _add_object(self, cls, **kw):
        # Instantiating an object will trigger potential validators:
        d = attr.asdict(cls(**kw))
        #
        # FIXME: check whether certain attributes should not be written to the table.
        #
        t = cls.__cldf_table__()
        for key in ['ID', 'Language_ID', 'Parameter_ID', 'Cognateset_ID']:
            # stringify/sluggify identifiers:
            if d.get(key) is not None:
                d[key] = '{0}'.format(d[key])
                if not ID_PATTERN.match(d[key]):
                    raise ValueError(
                        'invalid CLDF identifier {0}-{1}: {2}'.format(t, key, d[key]))
        if 'ID' not in d or d['ID'] not in self._obj_index[t]:
            if 'ID' in d:
                self._obj_index[t].add(d['ID'])
            self.objects[t].append(d)
        return d

    def add_cognate(self, lexeme=None, **kw):
        """
        Add a row to `CognateTable`.

        :param lexeme: An optional `FormTable` row (i.e. `dict`) to lookup lexeme data.
        :param kw: Attribute data to intialise a `self.dataset.cognate class` instance.
        :return: The (possibly augmented) instance data as `dict`.
        """
        if lexeme:
            if not isinstance(lexeme, collections.abc.Mapping):  # pragma: no cover
                raise TypeError('lexeme must be a mapping (`dict`) of lexeme attributes')
            kw.setdefault('Form_ID', lexeme['ID'])
            kw.setdefault('Form', lexeme['Form'])

        kw.setdefault('ID', self.cognate_id(kw))
        return self._add_object(self.dataset.cognate_class, **kw)

    def add_language(self, **kw):
        if (not getattr(self.args, 'dev', False)) and 'Glottocode' in kw \
                and hasattr(self.args, 'glottolog') \
                and kw['Glottocode'] in self.args.glottolog.api.cached_languoids:
            glang = self.args.glottolog.api.cached_languoids[kw['Glottocode']]
            for key, attribute in [
                ('Latitude', 'latitude'),
                ('Longitude', 'longitude'),
                ('Glottolog_Name', 'name'),
                ('ISO639P3code', 'iso')
            ]:
                if kw.get(key) is None:
                    kw[key] = getattr(glang, attribute)
            if kw.get('Family') is None:
                kw['Family'] = glang.lineage[0][0] if glang.lineage else glang.name
            if kw.get('Macroarea') is None:
                kw['Macroarea'] = glang.macroareas[0].name if glang.macroareas else None

        return self._add_object(self.dataset.language_class, **kw)

    def add_languages(self, id_factory='ID', lookup_factory=None):
        """
        Add languages as specified in a dataset's etc/languages.csv

        :param id_factory: A callable taking a dict describing a language as argument and returning\
        a value to be used as ID for the language or a `str` specifying a key in the language \
        `dict`.
        :param lookup_factory: A callable taking a language `dict` and returning a reverse \
        lookup key to associate the generated ID with (or a `str` specifying a key in the language \
        `dict`).
        :return: The `list` of language IDs which have been added, if no `lookup_factory` was \
        passed, otherwise an `OrderedDict`, mapping lookup to ID.
        """
        assert callable(id_factory) or isinstance(id_factory, str)
        ids = collections.OrderedDict()
        for i, kw in enumerate(self.dataset.languages):
            if (not kw.get('Glottocode')) and kw.get('ISO639P3code'):
                kw['Glottocode'] = self.dataset.glottolog.glottocode_by_iso.get(kw['ISO639P3code'])
            kw['ID'] = id_factory(kw) if callable(id_factory) else kw[id_factory]
            if lookup_factory is None:
                key = i
            else:
                key = lookup_factory(kw) if callable(lookup_factory) else kw[lookup_factory]
            ids[key] = kw['ID']
            self.add_language(**kw)
        return ids if lookup_factory else list(ids.values())

    def add_concept(self, **kw):
        if kw.get('Concepticon_ID'):
            gloss = self.dataset.concepticon.cached_glosses[int(kw['Concepticon_ID'])]
            if kw.get('Concepticon_Gloss') and kw.get('Concepticon_Gloss') != gloss:
                raise ValueError('Concepticon ID / Gloss mismatch %s != %s' % (
                    kw.get('Concepticon_Gloss'), gloss
                ))
            kw['Concepticon_Gloss'] = gloss
        return self._add_object(self.dataset.concept_class, **kw)

    def add_concepts(self, id_factory=lambda d: d.number, lookup_factory=None):
        """
        Add concepts as specified in a dataset's associated Concepticon concept list or in
        etc/concepts.csv

        :param id_factory: A callable taking a pyconcepticon.api.Concept object as argument and \
        returning a value to be used as ID for the concept.
        :param lookup_factory: A callable taking a `dict` object and returning a reverse \
        lookup key to associate the generated ID with (or a `str` specifying an attribute in \
        `Concept`).
        :return: The `list` of concept IDs which have been added, if no `lookup_factory` was \
        passed, otherwise an `OrderedDict`, mapping lookup to ID.
        """
        assert callable(id_factory) or isinstance(id_factory, str)

        ids, attrss = get_ids_and_attrs(
            # Read pyconcepticon.Concept instances either from a conceptlist in Concepticon, or from
            # etc/concepts.csv:
            get_concepts(self.dataset.conceptlists, self.dataset.concepts),
            {f.lower(): f for f in self.dataset.concept_class.fieldnames()},
            id_factory,
            lookup_factory,
        )
        for attrs in attrss:
            self.add_concept(**attrs)
        return ids if lookup_factory else list(ids.values())

    def align_cognates(self,
                       alm=None,
                       cognates=None,
                       column='Segments',
                       method='library'):
        from pylexibank.lingpy_util import iter_alignments

        iter_alignments(
            alm or self,
            cognates or self.objects['CognateTable'],
            column=column,
            method=method)



"""
Run makecldf command of a dataset
"""
import html

from clldutils import jsonlib
from cldfbench.cli_util import with_dataset, get_dataset
from cldfbench.commands import cldfreadme

from pylexibank.cli_util import add_catalogs, add_dataset_spec


def register(parser):
    add_dataset_spec(parser)
    add_catalogs(parser, with_clts=True)
    parser.add_argument('--verbose', action='store_true', default=False)
    parser.add_argument('--dev', action='store_true', default=False)


def run(args):
    dataset = get_dataset(args)
    dataset.concepticon = args.concepticon.api
    dataset.glottolog = args.glottolog.api
    with_dataset(args, 'makecldf', dataset=dataset)
    if not dataset.cldf_dir.joinpath('sources.bib').exists():
        args.log.warning(
            'The dataset has no sources at {0}'.format(dataset.cldf_dir.joinpath('sources.bib')))
    creators, contributors = dataset.get_creators_and_contributors(strict=False)

    def contrib(d):
        return {k: v for k, v in d.items() if k in {'name', 'affiliation', 'orcid', 'type'}}

    with jsonlib.update_ordered(dataset.dir / '.zenodo.json', indent=4) as md:
        md.update(
            {
                "title": dataset.metadata.title,
                "access_right": "open",
                "keywords": sorted(set(md.get("keywords", []) + ["linguistics", "cldf:Wordlist"])),
                "creators": [contrib(p) for p in creators],
                "contributors": [contrib(p) for p in contributors],
                "communities": [
                    {"identifier": community_id}
                    for community_id in sorted(
                        set([r["identifier"] for r in md.get("communities", [])] + ["lexibank"])
                    )
                ],
                "upload_type": "dataset",
            }
        )
        if dataset.metadata.citation:
            md['description'] = "<p>Cite the source of the dataset as:</p>\n\n" \
                                "<blockquote>\n<p>{}</p>\n</blockquote>".format(
                html.escape(dataset.metadata.citation))
        if dataset.metadata.zenodo_license:
            md['license'] = {'id': dataset.metadata.zenodo_license}
    cldfreadme.run(args)



from pathlib import Path
import argparse

from pylexibank.providers.tob import TOB

HTML = '<div class="results_record">' \
       '<div><span></span><span>1</span></div>' \
       '<div><span></span><span>concept</span></div>' \
       '<div>' \
       '<span>A</span>' \
       '<span class="Glottolog: abcd1234">Name</span>' \
       '<span>C</span>' \
       '<span></span>' \
       '<span>1</span>' \
       '</div>' \
       '</div>'


def test_TOB(tmp_path, mocker, concepticon, glottolog):
    class DS(TOB):
        dir = tmp_path
        id = 'test'
        name = 'name'
        dset = 'dset'

    tmp_path.joinpath('metadata.json').write_text('{"conceptlist": "Wang-2004-471"}', encoding='utf8')

    class Requests(mocker.Mock):
        def get(self, *args, **kw):
            return mocker.Mock(
                status_code=200,
                iter_content=mocker.Mock(return_value=[HTML.encode('utf8')]))

    ds = DS(concepticon=concepticon, glottolog=glottolog)
    mocker.patch('cldfbench.datadir.requests', Requests())
    ds._cmd_download(mocker.Mock())
    ds._cmd_makecldf(argparse.Namespace(verbose=False, log=None, dev=False))



"""
Check forms against a dataset's orthography profile.
"""
from clldutils.clilib import Table, add_format
from cldfbench.cli_util import with_dataset
from pylexibank.cli_util import add_dataset_spec


def register(parser):
    add_dataset_spec(parser)
    add_format(parser, default='pipe')


def run(args):
    with_dataset(args, check_segments)


def check_segments(dataset, args):
    with Table(args, 'Type', 'ID', 'Value', 'Form', 'Graphemes', 'Segments') as table:
        for row in sorted(
                dataset.cldf_dir.read_csv('forms.csv', dicts=True), key=lambda r: r['ID']):
            etype, tokens = None, row['Segments']
            tkl = row['Segments'].split()
            if tkl:
                if any(x in y for x, y in [
                    (tkl[0], ['+', '_', '#']),
                    (tkl[-1], ['+', '_', '#']),
                    ('+ +', tokens),
                    ('_ _', tokens),
                    ('#', tokens),
                    ('_', tokens),
                ]):
                    etype = 1
            else:
                etype = 2  # pragma: no cover
            if etype:
                table.append([
                    etype, row['ID'], row['Value'], row['Form'], row['Graphemes'], row['Segments']])



import re
import pathlib
import itertools
import collections

from termcolor import colored
from tqdm import tqdm

from clldutils.misc import slug
from clldutils.badge import Colors, badge
from clldutils import jsonlib
from cldfbench.datadir import get_url
from pycldf.sources import Source, Reference
from pyconcepticon.api import Concept

__all__ = ['progressbar', 'getEvoBibAsBibtex', 'iter_repl']
YEAR_PATTERN = re.compile(r'\s+\(?(?P<year>[1-9][0-9]{3}(-[0-9]+)?)(\)|\.)')


def progressbar(iterable=None, **kw):
    kw.setdefault('leave', False)
    kw.setdefault('desc', 'cldfbench')
    return tqdm(iterable=iterable, **kw)


# backwards compat:
pb = progressbar


def iter_repl(seq, subseq, repl):
    """
    Replace sub-list `subseq` in `seq` with `repl`.
    """
    seq, subseq, repl = list(seq), list(subseq), list(repl)
    subseq_len = len(subseq)
    rem = seq[:]
    while rem:
        if rem[:subseq_len] == subseq:
            for c in repl:
                yield c
            rem = rem[subseq_len:]
        else:
            yield rem.pop(0)


def split_by_year(s):
    match = YEAR_PATTERN.search(s)
    if match:
        return s[:match.start()].strip(), match.group('year'), s[match.end():].strip()
    return None, None, s


def get_reference(author, year, title, pages, sources, id_=None, genre='misc'):
    kw = {'title': title}
    id_ = id_ or None
    if author and year:
        id_ = id_ or slug(author + year)
        kw.update(author=author, year=year)
    elif title:
        id_ = id_ or slug(title)

    if not id_:
        return

    source = sources.get(id_)
    if source is None:
        sources[id_] = source = Source(genre, id_, **kw)

    return Reference(source, pages)


def get_badge(ratio, name):
    if ratio >= 0.99:
        color = Colors.brightgreen
    elif ratio >= 0.9:
        color = 'green'
    elif ratio >= 0.8:
        color = Colors.yellowgreen
    elif ratio >= 0.7:
        color = Colors.yellow
    elif ratio >= 0.6:
        color = Colors.orange
    else:
        color = Colors.red
    ratio = int(round(ratio * 100))
    return badge(name, '%s%%' % ratio, color, label="{0}: {1}%".format(name, ratio))


def sorted_obj(obj):
    res = obj
    if isinstance(obj, dict):
        res = collections.OrderedDict()
        obj.pop(None, None)
        for k, v in sorted(obj.items()):
            res[k] = sorted_obj(v)
    elif isinstance(obj, (list, set)):
        res = [sorted_obj(v) for v in obj]
    return res


def log_dump(fname, log=None):
    if log:
        log.info('file written: {0}'.format(colored(fname.as_posix(), 'green')))


def jsondump(obj, fname, log=None):
    fname = pathlib.Path(fname)
    if fname.exists():
        d = jsonlib.load(fname)
        d.update(obj)
        obj = d
    jsonlib.dump(sorted_obj(obj), fname, indent=4)
    log_dump(fname, log=log)
    return obj


def getEvoBibAsBibtex(*keys, **kw):
    """Download bibtex format and parse it from EvoBib"""
    res = []
    for key in keys:
        bib = get_url(
            "http://bibliography.lingpy.org/raw.php?key=" + key,
            log=kw.get('log')).text
        try:
            res.append('@' + bib.split('@')[1].split('</pre>')[0])
        except IndexError:  # pragma: no cover
            raise KeyError('Missing entry {0} in evobib'.format(key))
    return '\n\n'.join(res)


def get_concepts(conceptlists, concepts):
    """
    Read pyconcepticon.Concept instances either from a conceptlist in Concepticon, or from
    etc/concepts.csv:

    :param conceptlists:
    :param concepts:
    :return:
    """
    if conceptlists:
        return list(itertools.chain(*[cl.concepts.values() for cl in conceptlists]))

    res = []
    fields = Concept.public_fields()
    for i, concept in enumerate(concepts, start=1):
        kw, attrs = {}, {}
        for k, v in concept.items():
            if k.lower() in fields:
                kw[k.lower()] = v
            else:
                attrs[k.lower()] = v

        if not kw.get('id'):
            kw['id'] = str(i)
        if not kw.get('number'):
            kw['number'] = str(i)
        res.append(Concept(attributes=attrs, **kw))
    return res


def get_ids_and_attrs(concepts, fieldnames, id_factory, lookup_factory):
    id_lookup, objs = collections.OrderedDict(), []

    for i, c in enumerate(concepts):
        try:
            # `id_factory` might expect a pyconcepticon.Concept instance as input:
            id_ = id_factory(c) if callable(id_factory) else getattr(c, id_factory)
        except AttributeError:
            id_ = None
        attrs = dict(
            ID=id_,
            Name=c.label,
            Concepticon_ID=c.concepticon_id,
            Concepticon_Gloss=c.concepticon_gloss)
        for fl, f in fieldnames.items():
            if f not in attrs:
                if fl in c.attributes:
                    attrs[f] = c.attributes[fl]
                if hasattr(c, fl):
                    attrs[f] = getattr(c, fl)
        if attrs['ID'] is None:
            attrs['ID'] = id_factory(attrs) if callable(id_factory) else attrs[id_factory]
        if lookup_factory is None:
            key = i
        else:
            try:
                key = lookup_factory(attrs) if callable(lookup_factory) else attrs[lookup_factory]
            except (KeyError, AttributeError):
                key = lookup_factory(c) if callable(lookup_factory) else getattr(c, lookup_factory)
        id_lookup[key] = attrs['ID']
        objs.append(attrs)

    return id_lookup, objs



from pylexibank.providers.clld import CLLD


def test_CLLD(mocker, repos, tmp_path):
    class Response(mocker.Mock):
        def iter_content(self, *args, **kw):
            print(repos.joinpath('wold_dataset.cldf.zip'))
            yield repos.joinpath('wold_dataset.cldf.zip').read_bytes()
    mocker.patch('cldfbench.datadir.get_url', mocker.Mock(return_value=Response()))

    class WOLD(CLLD):
        id = 'wold'
        dir = tmp_path

    ds = WOLD()
    assert ds.url()
    ds._cmd_download(mocker.Mock())
    assert ds.raw_dir.exists()
    assert ds.raw_dir.glob('*.csv')
    assert ds.original_cldf
    with ds.cldf_writer(mocker.Mock()) as w:
        ds.add_sources(w)



import collections

import attr
import pyclts
import pyclts.models
from clldutils.markup import Table


@attr.s
class Analysis(object):
    # map segments to frequency
    segments = attr.ib(default=attr.Factory(collections.Counter))
    # aggregate segments which are invalid for lingpy
    bipa_errors = attr.ib(default=attr.Factory(set))
    # aggregate segments which are invalid for clpa
    sclass_errors = attr.ib(default=attr.Factory(set))
    # map clpa-replaceable segments to their replacements
    replacements = attr.ib(default=collections.defaultdict(set))
    # count number of errors
    general_errors = attr.ib(default=0)


@attr.s
class Stats(Analysis):
    inventory_size = attr.ib(default=0)
    invalid_words = attr.ib(default=attr.Factory(list))
    invalid_words_count = attr.ib(default=0)
    bad_words = attr.ib(default=attr.Factory(list))
    bad_words_count = attr.ib(default=0)


def valid_sequence(segments):
    """
    Make sure that a list of segments does not have any wrong segmentations.
    """
    if not ''.join(segments).strip():
        return False
    elif (
        '_' in segments or  # noqa: W504
        '#' in segments or  # noqa: W504
        segments[0] == "+" or  # noqa: W504
        segments[-1] == "+" or  # noqa: W504
        ("+" in segments and segments[segments.index("+") + 1] == "+")
    ):
        return False
    return segments


# Note: We use a mutable default argument intentionally to serve as a cache.
def analyze(clts, segments, analysis, lookup=dict(bipa={}, dolgo={})):
    """
    Test a sequence for compatibility with CLPA and LingPy.

    :param analysis: Pass a `TranscriptionAnalysis` instance for cumulative reporting.
    """
    # raise a ValueError in case of empty segments/strings
    if not segments:
        raise ValueError('Empty sequence.')

    # test if at least one element in `segments` has information
    # (helps to catch really badly formed input, such as ['\n']
    if not [segment for segment in segments if segment.strip()]:
        raise ValueError('No information in the sequence.')

    # build the phonologic and sound class analyses
    try:
        bipa_analysis, sc_analysis = [], []

        for s in segments:
            a = lookup['bipa'].get(s)
            if a is None:
                a = lookup['bipa'].setdefault(s, clts.bipa[s])
            bipa_analysis.append(a)

            sc = lookup['dolgo'].get(s)
            if sc is None:
                sc = lookup['dolgo'].setdefault(s, clts.bipa.translate(s, clts.soundclass('dolgo')))
            sc_analysis.append(sc)
    except:  # noqa; pragma: no cover
        print(segments)
        raise

    # compute general errors; this loop must take place outside the
    # following one because the code for computing single errors (either
    # in `bipa_analysis` or in `soundclass_analysis`) is unnecessary
    # complicated
    for sound_bipa, sound_class in zip(bipa_analysis, sc_analysis):
        if isinstance(sound_bipa, pyclts.models.UnknownSound) or sound_class == '?':
            analysis.general_errors += 1

    # iterate over the segments and analyses, updating counts of occurrences
    # and specific errors
    for segment, sound_bipa, sound_class in zip(segments, bipa_analysis, sc_analysis):
        # update the segment count
        analysis.segments.update([segment])

        # add an error if we got an unknown sound, otherwise just append
        # the `replacements` dictionary
        if isinstance(sound_bipa, pyclts.models.UnknownSound):
            analysis.bipa_errors.add(segment)
        else:
            analysis.replacements[sound_bipa.source].add(str(sound_bipa))

        # update sound class errors, if any
        if sound_class == '?':
            analysis.sclass_errors.add(segment)

    return segments, bipa_analysis, sc_analysis, analysis


TEMPLATE = """
# Detailed transcription record

## Segments

{0}

## Unsegmentable lexemes (up to 100 only)

{1}

## Words with invalid segments (up to 100 only)

{2}
"""


def report(analysis):
    segments = Table('Segment', 'Occurrence', 'BIPA', 'CLTS SoundClass')
    for a, b in sorted(
            analysis['stats']['segments'].items(), key=lambda x: (-x[1], x[0])):
        c, d = '✓', '✓'
        if a in analysis['stats']['sclass_errors']:
            c = '✓' if a not in analysis['stats']['bipa_errors'] else '?'
            d = ', '.join(analysis['stats']['sclass_errors'][a]) \
                if a not in analysis['stats']['sclass_errors'] else '?'

        # escape pipe for markdown table if necessary
        a = a.replace('|', '&#124;')

        segments.append([a, b, c, d])

    invalid = Table('ID', 'LANGUAGE', 'CONCEPT', 'FORM')
    for row in analysis['stats']['invalid_words']:
        invalid.append(row)  # pragma: no cover

    words = Table('ID', 'LANGUAGE', 'CONCEPT', 'FORM', 'SEGMENTS')
    for row in analysis['stats']['bad_words']:
        words.append(row)
    return TEMPLATE.format(
        segments.render(verbose=True),
        invalid.render(verbose=True),
        words.render(verbose=True))



import pytest

from pylexibank.metadata import *


def test_conceptlist():
    md = LexibankMetadata(conceptlist='Swadesh-1964-100', related='abc')
    assert isinstance(md.conceptlist, list)
    assert 'concepticon.clld.org' in md.markdown()


STANDARD_TITLE = """CLDF dataset derived from A et al.'s "Paper" from 1934"""


@pytest.mark.parametrize('title', [
    "",
    STANDARD_TITLE.replace('CLDF', 'cldf'),
    STANDARD_TITLE.replace("et al.'s", 'et al.’s'),
    STANDARD_TITLE.replace('"', "'"),
    STANDARD_TITLE.replace('1934', '34'),
])
def test_invalid_standard_title(title):
    with pytest.raises(AssertionError):
        check_standard_title(title)


def test_valid_standard_title():
    assert check_standard_title(STANDARD_TITLE) is None
    assert check_standard_title(STANDARD_TITLE.replace(" et al.'s", "brams'")) is None


def test_get_creators_and_contributors():
    assert len(get_creators_and_contributors([])[0]) == 0
    cr, co = get_creators_and_contributors(['a|role', '---|---', 'c|staff'], strict=False)
    assert len(cr) == 0 and len(co) == 1

    with pytest.raises(KeyError):
        get_creators_and_contributors(['a|role', '---|---', 'c|staff'], strict=True)

    assert len(get_creators_and_contributors(['a|role', '---|---', 'c|author'])[0]) == 1
    assert len(get_creators_and_contributors(['a|role', '---|---', 'c|author', 'a', 'e|f'])[0]) == 1

    with pytest.raises(AssertionError):
        get_creators_and_contributors(['---|---', 'c|d'])



import logging
import pathlib

from pylexibank.profile import Profile


def test_init():
    prf = Profile({'Grapheme': 'x', 'IPA': 'y'})
    assert prf.graphemes['^']['IPA'] is None


def test_sort(clts):
    prf = Profile({'Grapheme': 'ab', 'IPA': 'z'}, {'Grapheme': 'x', 'IPA': 'y'})
    assert list(prf.graphemes.keys())[0] == 'ab'
    prf.sort()
    assert list(prf.graphemes.keys())[0] == '^'
    assert list(prf.graphemes.keys())[-1] == 'ab'

    prf.sort(clts=clts)
    assert list(prf.graphemes.keys())[-1] == 'x'


def test_trim():
    prf = Profile(
        {'Grapheme': 'ab', 'IPA': 'x y'},
        {'Grapheme': 'a', 'IPA': 'x'},
        {'Grapheme': 'b', 'IPA': 'y'},
    )
    assert prf.trim() == 1
    assert 'ab' not in prf.graphemes


def test_augment(clts):
    prf = Profile(
        {'Grapheme': '^a', 'IPA': 'z'},
        {'Grapheme': 'a', 'IPA': 'x'},
        {'Grapheme': 'b', 'IPA': 'y'},
    )
    prf.augment(['aab', 'ba', 'aba'])
    assert prf.graphemes['^a']['FREQUENCY'] == 2
    assert prf.graphemes['a']['FREQUENCY'] == 3

    prf.augment(['aab', 'ba', 'aba'], clts=clts)
    assert prf.graphemes['^a']['SCA'] == 'S'


def test_write(tmp_path):
    fname = tmp_path / 'profile.tsv'
    prf = Profile({'Grapheme': 'ab', 'IPA': 'z'}, {'Grapheme': 'x', 'IPA': 'y'})
    prf.write(fname)
    assert Profile.from_file(fname).graphemes == prf.graphemes


def test_clean(clts):
    prf = Profile(
        {'Grapheme': 'a', 'IPA': 'ƛ', 'CODEPOINTS': ''},
        {'Grapheme': 'b', 'IPA': 'b/ƛ', 'CODEPOINTS': ''},
    )
    prf.clean(clts)
    assert prf.graphemes['a']['IPA'] != 'ƛ'
    assert prf.graphemes['a']['CODEPOINTS'] == 'U+0061'
    assert prf.graphemes['b']['IPA'] == 'b/tɬ'


def test_check(caplog, tmp_path, clts):
    prf_path = tmp_path / 'profile.tsv'

    prf_path.write_text('Grapheme\tIPA\na\tx\na\tx\n')
    prf = Profile.from_file(prf_path)
    prf.check(log=logging.getLogger(__name__))
    assert caplog.records[-1].levelname == 'WARNING'

    prf_path.write_text('Grapheme\tIPA\na\tx\na\ty\n')
    prf.check(log=logging.getLogger(__name__))
    assert caplog.records[-1].levelname == 'ERROR'

    prf_path.write_text('Grapheme\tIPA\na\t°\n')
    prf.check(clts=clts, log=logging.getLogger(__name__))
    assert caplog.records[-1].levelname == 'ERROR'



"""
Check forms against a dataset's orthography profile.
"""
import re
import collections
from unicodedata import normalize

from cldfbench.cli_util import with_dataset, add_catalog_spec
from clldutils.clilib import Table, add_format
from pylexibank.cli_util import add_dataset_spec


def normalized(string, mode="NFD"):
    return normalize(mode, string)


def register(parser):
    add_format(parser, default="pipe")
    add_dataset_spec(parser)
    add_catalog_spec(parser, "clts")
    parser.add_argument(
        '--language',
        help="Select a language",
        default=None)
    parser.add_argument(
        "--noprofile",
        help="Ignore profile use segments",
        action="store_true",
        default=False)


def codepoints(string):
    return " ".join(["U+" + x.rjust(4, "0") for x in [hex(ord(char))[2:] for char in string]])


def run(args):
    with_dataset(args, check_profile)


def check_profile(dataset, args):
    def n(tk):
        return re.sub('>>$', '', re.sub(r'^<<', '', tk))
    visited = {}
    problems = collections.OrderedDict([
        ('generated', collections.defaultdict(list)),
        ('modified', collections.defaultdict(list)),
        ('slashed', collections.defaultdict(list)),
        ('unknown', collections.defaultdict(list)),
        ('missing', collections.defaultdict(list)),
    ])
    for row in dataset.cldf_dir.read_csv("forms.csv", dicts=True):
        if not args.language or args.language == row["Language_ID"]:
            kw = dict(column="IPA")
            if 'Profile' in row:  # A multi-profile dataset.
                kw['profile'] = None if row['Profile'] == 'default' else row['Profile']
            tokens = [normalized(t) for t in (
                dataset.tokenizer(row, row["Form"], **kw)
                if dataset.tokenizer and not args.noprofile
                else row["Segments"].split()
            )]
            for tk in set(tokens):
                if tk not in visited:
                    sound = args.clts.api.bipa[tk]
                    if tk.startswith("<<") and tk.endswith(">>"):
                        visited[tk] = "missing"
                    elif sound.type == "unknownsound":
                        visited[tk] = "unknown"
                    elif sound.generated:
                        visited[tk] = "generated"
                    elif str(sound) not in {tk, normalized(tk), normalized(tk, mode='NFC')}:
                        if "/" in tk and str(sound) == tk.split('/')[1]:
                            visited[tk] = "slashed"
                        else:
                            visited[tk] = "modified"
                if tk in visited:
                    problems[visited[tk]][n(tk)].append(
                        (" ".join(tokens), row["Form"], row["Graphemes"]))

    with Table(
        args,
        'Category',
        'Grapheme',
        'Grapheme-UC',
        'Non-normal',
        'Diacritics',
        'BIPA',
        "BIPA-UC",
        "Segments",
        "Graphemes",
        "Count",
    ) as table:
        for cat, tokens in problems.items():
            for tk, values in tokens.items():
                table.append([
                    cat,
                    tk,
                    codepoints(tk.split('/')[0] if cat == 'slashed' else tk),
                    "*" if cat == 'generated' and tk != str(args.clts.api.bipa[tk]) else "",
                    "◌" + tk if cat in {'missing', 'unknown'} else '',
                    str(args.clts.api.bipa[tk]),
                    args.clts.api.bipa[tk].codepoints,
                    values[0][0],
                    values[0][1],
                    len(values),
                ])



"""
Load dataset(s) into the lexibank SQLite database.
"""
from cldfbench.cli_util import with_datasets

from pylexibank.cli_util import add_db, get_db, add_dataset_spec


def register(parser):
    add_db(parser)
    add_dataset_spec(parser, multiple=True)


def run(args):
    db = get_db(args)
    with_datasets(args, db.unload)



import shutil

import pytest

from pylexibank.providers import abvd


@pytest.fixture
def abvd_dataset(repos, tmp_path, glottolog, concepticon):
    shutil.copytree(repos / 'datasets' / 'abvd', tmp_path / 'abvd')

    class Dataset(abvd.BVD):
        id = 'x'
        SECTION = 'y'
        dir = tmp_path / 'abvd'

    return Dataset(glottolog=glottolog, concepticon=concepticon)


def test_Wordlist(abvd_dataset, mocker):
    with abvd_dataset.cldf_writer(mocker.MagicMock()) as ds:
        for wl in abvd_dataset.iter_wordlists():
            wl.to_cldf(ds, {})
            assert wl.name
            assert wl.id





"""Functionality to load a set of CLDF datasets into a sqlite db.

Notes:
- Only CLDF components will be loaded into the db.
- The names of the columns in the database are the names from the CSV files, not the
  preferred labels for the corresponding CLDF properties.
"""
import json
import pathlib
import sqlite3
import contextlib
import collections

import attr
from csvw.datatypes import DATATYPES
from clldutils.misc import nfilter
from clldutils.jsonlib import load

from pycldf.terms import term_uri
from pycldf.sources import Sources


def identity(s):
    return s


TYPE_MAP = {
    'string': ('TEXT', identity),
    'integer': ('INTEGER', identity),
    'boolean': ('INTEGER', lambda s: s if s is None else int(s)),
    'decimal': ('REAL', lambda s: s if s is None else float(s)),
}
BIBTEX_FIELDS = [
    'address',  # Publisher's address
    'annote',  # An annotation for annotated bibliography styles (not typical)
    'author',  # The name(s) of the author(s) (separated by and)
    'booktitle',  # The title of the book, if only part of it is being cited
    'chapter',  # The chapter number
    'crossref',  # The key of the cross-referenced entry
    'edition',  # The edition of a book, long form (such as "First" or "Second")
    'editor',  # The name(s) of the editor(s)
    'eprint',  # A specification of electronic publication, preprint or technical report
    'howpublished',  # How it was published, if the publishing method is nonstandard
    'institution',  # institution involved in the publishing,not necessarily the publisher
    'journal',  # The journal or magazine the work was published in
    'key',  # A hidden field used for specifying or overriding the orderalphabetical order
    'month',  # The month of publication (or, if unpublished, the month of creation)
    'note',  # Miscellaneous extra information
    'number',  # The "(issue) number" of a journal, magazine, or tech-report
    'organization',  # The conference sponsor
    'pages',  # Page numbers, separated either by commas or double-hyphens.
    'publisher',  # The publisher's name
    'school',  # The school where the thesis was written
    'series',  # The series of books the book was published in
    'title',  # The title of the work
    'type',  # The field overriding the default type of publication
    'url',  # The WWW address
    'volume',  # The volume of a journal or multi-volume book
    'year',
]


PROPERTY_URL_TO_COL = collections.defaultdict(dict)
for table in load(pathlib.Path(__file__).parent / 'cldf-metadata.json')['tables']:
    for col in table['tableSchema']['columns']:
        if col.get('propertyUrl'):
            PROPERTY_URL_TO_COL[table['dc:conformsTo'].split('#')[1]][col['propertyUrl']] = \
                col['name']


def insert(db, table, keys, *rows, **kw):
    if rows:
        if isinstance(keys, str):
            keys = [k.strip() for k in keys.split(',')]
        sql = "INSERT INTO {0} ({1}) VALUES ({2})".format(
            table, ','.join(keys), ','.join(['?' for _ in keys]))
        db.executemany(sql, rows)


def quoted(*names):
    return ','.join('`{0}`'.format(name) for name in names)


@attr.s
class ColSpec(object):
    """
    A `ColSpec` captures sufficient information about a `Column` for the DB schema.
    """
    name = attr.ib()
    csvw_type = attr.ib(default='string', converter=lambda s: s if s else 'string')
    separator = attr.ib(default=None)
    primary_key = attr.ib(default=None)
    db_type = attr.ib(default=None)
    convert = attr.ib(default=None)
    cldf_name = attr.ib(default=None)

    def __attrs_post_init__(self):
        if self.csvw_type in TYPE_MAP:
            self.db_type, self.convert = TYPE_MAP[self.csvw_type]
        else:
            self.db_type = 'TEXT'
            self.convert = DATATYPES[self.csvw_type].to_string
        if not self.cldf_name:
            self.cldf_name = self.name

    @property
    def sql(self):
        return '`{0.name}` {0.db_type}'.format(self)


@attr.s
class TableSpec(object):
    """
    A `TableSpec` captures sufficient information about a `Table` for the DB schema.
    """
    name = attr.ib()
    columns = attr.ib(default=attr.Factory(list))
    foreign_keys = attr.ib(default=attr.Factory(list))
    consumes = attr.ib(default=None)
    primary_key = attr.ib(default=None)

    @property
    def sql(self):
        clauses = [col.sql for col in self.columns]  # limit to non-local columns!
        clauses.append('`dataset_ID` TEXT NOT NULL')
        if self.primary_key:
            clauses.append('PRIMARY KEY(`dataset_ID`, `{0}`)'.format(self.primary_key))
        clauses.append('FOREIGN KEY(`dataset_ID`) REFERENCES dataset(`ID`) ON DELETE CASCADE')
        for fk, ref, refcols in self.foreign_keys:
            clauses.append('FOREIGN KEY({0}) REFERENCES {1}({2}) ON DELETE CASCADE'.format(
                quoted(*fk), ref, quoted(*refcols)))
        return "CREATE TABLE {0} (\n    {1}\n)".format(self.name, ',\n    '.join(clauses))


def schema(ds):
    """
    Convert the table and column descriptions of a `Dataset` into specifications for the
    DB schema.

    :param ds:
    :return: A pair (tables, reference_tables).
    """
    tables, ref_tables = {}, {}
    table_lookup = {t.url.string: t for t in ds.tables if ds.get_tabletype(t)}
    for table in table_lookup.values():
        spec = TableSpec(ds.get_tabletype(table))
        spec.primary_key = [
            c for c in table.tableSchema.columns if
            c.propertyUrl and c.propertyUrl.uri == term_uri('id')][0].name
        # Map the column name to the default:
        if spec.name in PROPERTY_URL_TO_COL:
            spec.primary_key = PROPERTY_URL_TO_COL[spec.name][term_uri('id')]
        for c in table.tableSchema.columns:
            if c.propertyUrl and c.propertyUrl.uri == term_uri('source'):
                # A column referencing sources is replaced by an association table.
                otype = ds.get_tabletype(table).replace('Table', '')
                ref_tables[ds.get_tabletype(table)] = TableSpec(
                    '{0}Source'.format(otype),  # The name of the association table.
                    [ColSpec(otype + '_ID'), ColSpec('Source_ID'), ColSpec('Context')],
                    [
                        (  # The foreign key to the referencing object:
                            ['dataset_ID', otype + '_ID'],
                            ds.get_tabletype(table),
                            ['dataset_ID', spec.primary_key]),
                        (  # The foreign key to the referenced source:
                            ['dataset_ID', 'Source_ID'],
                            'SourceTable',
                            ['dataset_ID', 'ID']),
                    ],
                    c.name)
            else:
                cname = c.header
                if c.propertyUrl and spec.name in PROPERTY_URL_TO_COL:
                    if c.propertyUrl.uri in PROPERTY_URL_TO_COL[spec.name]:
                        cname = PROPERTY_URL_TO_COL[spec.name][c.propertyUrl.uri]
                spec.columns.append(ColSpec(
                    cname,
                    c.datatype.base if c.datatype else c.datatype,
                    c.separator,
                    cname == spec.primary_key,
                    cldf_name=c.header))
        listvalued = set(c.name for c in table.tableSchema.columns if c.separator)
        for fk in table.tableSchema.foreignKeys:
            if fk.reference.schemaReference:
                # We only support Foreign Key references between tables!
                continue  # pragma: no cover
            ref = table_lookup[fk.reference.resource.string]
            ref_type = ds.get_tabletype(ref)
            if ref_type:
                colRefs = sorted(fk.columnReference)
                if any(c in listvalued for c in colRefs):
                    # We drop list-valued foreign keys
                    continue  # pragma: no cover
                if spec.name in PROPERTY_URL_TO_COL:  # Current table is a standard CLDF component.
                    # Must map foreign keys
                    colRefs = []
                    for c in sorted(fk.columnReference):
                        c = ds[spec.name, c]
                        if c.propertyUrl and c.propertyUrl.uri in PROPERTY_URL_TO_COL[spec.name]:
                            colRefs.append(PROPERTY_URL_TO_COL[spec.name][c.propertyUrl.uri])
                        else:
                            colRefs.append(c.header)  # pragma: no cover
                rcolRefs = sorted(fk.reference.columnReference)
                if ref_type in PROPERTY_URL_TO_COL:
                    # Must map foreign key targets!
                    rcolRefs = []
                    for c in sorted(fk.reference.columnReference):
                        c = ds[ref_type, c]
                        if c.propertyUrl and c.propertyUrl.uri in PROPERTY_URL_TO_COL[ref_type]:
                            rcolRefs.append(PROPERTY_URL_TO_COL[ref_type][c.propertyUrl.uri])
                        else:
                            rcolRefs.append(c.header)  # pragma: no cover
                spec.foreign_keys.append((
                    tuple(['dataset_ID'] + colRefs),
                    ds.get_tabletype(table_lookup[fk.reference.resource.string]),
                    tuple(['dataset_ID'] + rcolRefs)))
        tables[spec.name] = spec

    # must determine the order in which tables must be created!
    ordered = collections.OrderedDict()
    i = 0
    #
    # We loop through the tables repeatedly, and whenever we find one, which has all
    # referenced tables already in ordered, we move it from tables to ordered.
    #
    while tables and i < 100:
        i += 1
        for table in list(tables.keys()):
            if all(ref[1] in ordered for ref in tables[table].foreign_keys):
                # All referenced tables are already created.
                ordered[table] = tables.pop(table)
                break
    if tables:  # pragma: no cover
        raise ValueError('there seem to be cyclic dependencies between the tables')

    return list(ordered.values()), ref_tables


class Database(object):
    def __init__(self, fname):
        """
        A `Database` instance is initialized with a file path.

        :param fname: Path to a file in the file system where the db is to be stored.
        """
        self.fname = pathlib.Path(fname)

    def drop(self):
        if self.fname.exists():
            self.fname.unlink()

    def connection(self):
        return contextlib.closing(sqlite3.connect(self.fname.as_posix()))

    def create(self, force=False, exists_ok=False):
        """
        Creates a db file with the core schema.

        :param force: If `True` an existing db file will be overwritten.
        """
        if self.fname and self.fname.exists():
            if force:
                self.drop()
            elif exists_ok:
                return
            else:
                raise ValueError('db file already exists, use force=True to overwrite')
        with self.connection() as db:
            db.execute(
                """\
CREATE TABLE dataset (
    ID TEXT PRIMARY KEY NOT NULL,
    name TEXT,
    version TEXT,
    metadata_json TEXT
)""")
            db.execute("""\
CREATE TABLE datasetmeta (
    dataset_ID TEXT ,
    key TEXT,
    value TEXT,
    PRIMARY KEY (dataset_ID, key),
    FOREIGN KEY(dataset_ID) REFERENCES dataset(ID)
)""")
            db.execute("""\
CREATE TABLE SourceTable (
    dataset_ID TEXT ,
    ID TEXT ,
    bibtex_type TEXT,
    {0}
    extra TEXT,
    PRIMARY KEY (dataset_ID, ID),
    FOREIGN KEY(dataset_ID) REFERENCES dataset(ID)
)""".format('\n    '.join('`{0}` TEXT,'.format(f) for f in BIBTEX_FIELDS)))

    def fetchone(self, sql, params=None, conn=None, verbose=False):
        return self._fetch(sql, 'fetchone', params, conn, verbose=verbose)

    def fetchall(self, sql, params=None, conn=None, verbose=False):
        return self._fetch(sql, 'fetchall', params, conn, verbose=verbose)

    def _fetch(self, sql, method, params, conn, verbose=False):
        sql = self.sql.get(sql, sql)

        def _do(conn, sql, method):
            cu = conn.cursor()
            if verbose:
                print(sql)
            cu.execute(sql, params or ())
            return getattr(cu, method)()

        if not conn:
            with self.connection() as conn:
                return _do(conn, sql, method)
        else:
            return _do(conn, sql, method)

    @property
    def tables(self):
        res = {r[0]: {} for r in self.fetchall(
            "SELECT name FROM sqlite_master WHERE type='table'")}
        for t in res:
            res[t] = {r[1]: r[2] for r in self.fetchall(
                "PRAGMA table_info({0})".format(t))}
        return res

    def unload(self, dataset_id, args=None):
        dataset_id = getattr(dataset_id, 'id', dataset_id)
        with self.connection() as db:
            for table in self.tables:
                if table != 'dataset':
                    db.execute(
                        "DELETE FROM {0} WHERE dataset_ID = ?".format(table),
                        (dataset_id,))
            db.execute("DELETE FROM dataset WHERE ID = ?", (dataset_id,))
            db.commit()

    def _create_table_if_not_exists(self, table):
        if table.name in self.tables:
            return False

        with self.connection() as conn:
            conn.execute(table.sql)
        return True

    def load(self, ds, args=None, verbose=False):
        """
        Load a CLDF dataset into the database.

        :param dataset:
        :return:
        """
        print(ds)
        try:
            self.fetchone('select ID from dataset')
        except sqlite3.OperationalError:
            self.create(force=True)
        self.unload(ds)

        dataset = ds.cldf_reader()
        tables, ref_tables = schema(dataset)

        # update the DB schema:
        for t in tables:
            if self._create_table_if_not_exists(t):
                continue
            db_cols = {k.lower(): v for k, v in self.tables[t.name].items()}
            for col in t.columns:
                if col.name.lower() not in db_cols:
                    with self.connection() as conn:
                        conn.execute(
                            "ALTER TABLE {0} ADD COLUMN `{1.name}` {1.db_type}".format(
                                t.name, col))
                else:
                    if db_cols[col.name.lower()] != col.db_type:
                        raise ValueError(
                            'column {0}:{1} {2} redefined with new type {3}'.format(
                                t.name, col.name, db_cols[col.name.lower()], col.db_type))

        for t in ref_tables.values():
            self._create_table_if_not_exists(t)

        self.update_schema()

        # then load the data:
        with self.connection() as db:
            db.execute('PRAGMA foreign_keys = ON;')
            insert(
                db,
                'dataset',
                'ID,name,version,metadata_json',
                (
                    ds.id,
                    '{0}'.format(dataset),
                    ds.repo.hash() if ds.repo else '',
                    json.dumps(dataset.metadata_dict)))
            insert(
                db,
                'datasetmeta',
                'dataset_ID,key,value',
                *[(ds.id, k, '{0}'.format(v)) for k, v in dataset.properties.items()])

            # load sources:
            rows = []
            for src in dataset.sources.items():
                values = [ds.id, src.id, src.genre] + [src.get(k) for k in BIBTEX_FIELDS]
                values.append(
                    json.dumps({k: v for k, v in src.items() if k not in BIBTEX_FIELDS}))
                rows.append(tuple(values))
            insert(
                db,
                'SourceTable',
                ['dataset_ID', 'ID', 'bibtex_type'] + BIBTEX_FIELDS + ['extra'],
                *rows)

            # For regular tables, we extract and keep references to sources.
            refs = collections.defaultdict(list)

            for t in tables:
                # We want to lookup columns by the name used in the CLDF dataset.
                cols = {col.cldf_name: col for col in t.columns}
                # But we also want to look up primary keys by the database column name.
                cols_by_name = {col.name: col for col in t.columns}

                ref_table = ref_tables.get(t.name)
                rows, keys = [], []
                try:
                    for row in dataset[t.name]:
                        keys, values = ['dataset_ID'], [ds.id]
                        for k, v in row.items():
                            if ref_table and k == ref_table.consumes:
                                col = cols_by_name[t.primary_key]
                                refs[ref_table.name].append((row[col.cldf_name], v))
                            else:
                                col = cols[k]
                                if isinstance(v, list):
                                    v = (col.separator or ';').join(
                                        nfilter(col.convert(vv) for vv in v))
                                else:
                                    v = col.convert(v)
                                # FIXME: only if non-local!
                                keys.append("`{0}`".format(col.name))
                                values.append(v)
                        keys, values = self.update_row(t.name, keys, values)
                        rows.append(tuple(values))
                    insert(db, t.name, keys, *rows, **{'verbose': verbose})
                except FileNotFoundError:  # pragma: no cover
                    if t.name != 'CognateTable':  # An empty CognateTable is allowed.
                        raise  # pragma: no cover

            # Now insert the references, i.e. the associations with sources:
            for tname, items in refs.items():
                rows = []
                for oid, sources in items:
                    for source in sources:
                        sid, context = Sources.parse(source)
                        rows.append([ds.id, oid, sid, context])
                oid_col = '{0}_ID'.format(tname.replace('Source', ''))
                insert(db, tname, ['dataset_ID', oid_col, 'Source_ID', 'Context'], *rows)
            db.commit()

    def update_schema(self):
        for tname, cname, type_ in [
            ('ParameterTable', 'Ontological_Category', 'TEXT'),
            ('ParameterTable', 'Semantic_Field', 'TEXT'),
            ('LanguageTable', 'Latitude', 'REAL'),
            ('LanguageTable', 'Longitude', 'REAL'),
        ]:
            if cname not in self.tables[tname]:
                with self.connection() as conn:
                    conn.execute("ALTER TABLE {0} ADD COLUMN `{1}` {2}".format(
                        tname, cname, type_))

    def update_row(self, table, keys, values):
        return keys, values

    def load_concepticon_data(self, concepticon):
        conceptsets = []
        for csid in self.fetchall("SELECT distinct concepticon_id FROM parametertable"):
            cs = concepticon.conceptsets.get(csid[0])
            if cs:
                conceptsets.append((
                    cs.gloss, cs.ontological_category, cs.semanticfield, cs.id))

        with self.connection() as db:
            db.executemany(
                """\
UPDATE parametertable
SET concepticon_gloss = ?, ontological_category = ?, semantic_field = ?
WHERE concepticon_id = ?""",
                conceptsets)
            db.commit()

    def load_glottolog_data(self, glottolog):
        langs = []
        languoids = {lang.id: lang for lang in glottolog.languoids()}
        for gc in self.fetchall("SELECT distinct glottocode FROM languagetable"):
            lang = languoids.get(gc[0])
            if lang:
                langs.append((
                    lang.lineage[0][0] if lang.lineage else lang.name,
                    lang.macroareas[0].name if lang.macroareas else None,
                    lang.latitude,
                    lang.longitude,
                    lang.id))

        with self.connection() as db:
            db.executemany(
                """\
UPDATE languagetable
SET family = ?, macroarea = ?, latitude = ?, longitude = ?
WHERE glottocode = ?""",
                langs)
            db.commit()

    sql = {
        "conceptsets_by_dataset":
            "SELECT ds.id, count(distinct p.concepticon_id) "
            "FROM dataset as ds, parametertable as p "
            "WHERE ds.id = p.dataset_id GROUP BY ds.id",
        "families_by_dataset":
            "SELECT ds.id, count(distinct l.family) "
            "FROM dataset as ds, languagetable as l "
            "WHERE ds.id = l.dataset_id GROUP BY ds.id",
        "macroareas_by_dataset":
            "SELECT ds.id, group_concat(distinct l.macroarea) "
            "FROM dataset as ds, languagetable as l "
            "WHERE ds.id = l.dataset_id GROUP BY ds.id",
        "glottocodes_by_dataset":
            "SELECT ds.id, count(distinct l.glottocode) "
            "FROM dataset as ds, languagetable as l "
            "WHERE ds.id = l.dataset_id GROUP BY ds.id",
        "mapped_lexemes_by_dataset":
            "SELECT ds.id, count(distinct f.ID) "
            "FROM dataset as ds, formtable as f, languagetable as l, parametertable as p "
            "WHERE ds.id = f.dataset_id and f.Language_ID = l.ID and "
            "f.Parameter_ID = p.ID and l.glottocode is not null and "
            "p.concepticon_id is not null "
            "GROUP BY ds.id",
        "lexemes_by_dataset":
            "SELECT ds.id, count(f.ID) FROM dataset as ds, formtable as f "
            "WHERE ds.id = f.dataset_id GROUP BY ds.id",
    }



import re
import collections

import attr
from clldutils.misc import nfilter
from cldfbench.metadata import Metadata

__all__ = ['LexibankMetadata', 'check_standard_title', 'get_creators_and_contributors']

STANDARD_TITLE_PATTERN = re.compile(
    r'CLDF dataset derived from\s+'
    r'(?P<authors>[^"]+)'
    r'"(?P<title>[^"]+)"\s+from\s+'
    r'(?P<year>[0-9]{4})'
)
CONTRIBUTOR_TYPES = {
    'ContactPerson',
    'DataCollector',
    'DataCurator',
    'DataManager',
    'Distributor',
    'Editor',
    'Funder',
    'HostingInstitution',
    'Producer',
    'ProjectLeader',
    'ProjectManager',
    'ProjectMember',
    'RegistrationAgency',
    'RegistrationAuthority',
    'RelatedPerson',
    'Researcher',
    'ResearchGroup',
    'RightsHolder',
    'Supervisor',
    'Sponsor',
    'WorkPackageLeader',
    'Other',
}
LICENSES = {
    "AAL",
    "ADSL",
    "AFL-1.1",
    "AFL-3.0",
    "AGPL-1.0-only",
    "AGPL-3.0",
    "AGPL-3.0-only",
    "AGPL-3.0-or-later",
    "AMDPLPA",
    "AML",
    "AMPAS",
    "ANTLR-PD",
    "APL-1.0",
    "APSL-1.0",
    "APSL-1.1",
    "APSL-1.2",
    "APSL-2.0",
    "Adobe-2006",
    "Against-DRM",
    "Aladdin",
    "Apache-1.0",
    "Apache-1.1",
    "Apache-2.0",
    "Artistic-1.0",
    "Artistic-1.0-Perl",
    "Artistic-1.0-cl8",
    "Artistic-2.0",
    "BSD-1-Clause",
    "BSD-2-Clause",
    "BSD-2-Clause-FreeBSD",
    "BSD-3-Clause",
    "BSD-3-Clause-Clear",
    "BSD-3-Clause-LBNL",
    "BSD-3-Clause-No-Nuclear-License",
    "BSD-3-Clause-No-Nuclear-License-2014",
    "BSD-4-Clause",
    "BSD-4-Clause-UC",
    "BSD-Source-Code",
    "BSL-1.0",
    "Bahyph",
    "Barr",
    "Beerware",
    "BitTorrent-1.0",
    "BitTorrent-1.1",
    "CATOSL-1.1",
    "CC-BY-1.0",
    "CC-BY-3.0",
    "CC-BY-4.0",
    "CC-BY-NC-1.0",
    "CC-BY-NC-2.5",
    "CC-BY-NC-3.0",
    "CC-BY-NC-4.0",
    "CC-BY-NC-ND-1.0",
    "CC-BY-NC-ND-2.0",
    "CC-BY-NC-ND-2.5",
    "CC-BY-NC-ND-3.0",
    "CC-BY-NC-ND-4.0",
    "CC-BY-NC-SA-1.0",
    "CC-BY-NC-SA-3.0",
    "CC-BY-NC-SA-4.0",
    "CC-BY-ND-1.0",
    "CC-BY-ND-2.0",
    "CC-BY-ND-2.5",
    "CC-BY-ND-4.0",
    "CC-BY-SA-2.0",
    "CC-BY-SA-2.5",
    "CC-BY-SA-3.0",
    "CC-BY-SA-4.0",
    "CC0-1.0",
    "CDDL-1.0",
    "CDLA-Permissive-1.0",
    "CDLA-Sharing-1.0",
    "CECILL-1.1",
    "CECILL-2.0",
    "CECILL-2.1",
    "CECILL-B",
    "CECILL-C",
    "CNRI-Jython",
    "CNRI-Python",
    "CNRI-Python-GPL-Compatible",
    "CPAL-1.0",
    "CPOL-1.02",
    "CUA-OPL-1.0",
    "Caldera",
    "ClArtistic",
    "Condor-1.1",
    "CrystalStacker",
    "Cube",
    "D-FSL-1.0",
    "DSDP",
    "DSL",
    "ECL-2.0",
    "EFL-1.0",
    "EFL-2.0",
    "EPL-1.0",
    "EUDatagrid",
    "EUPL-1.0",
    "EUPL-1.1",
    "EUPL-1.2",
    "Entessa",
    "ErlPL-1.1",
    "Eurosym",
    "FAL-1.3",
    "FSFAP",
    "Fair",
    "Frameworx-1.0",
    "GFDL-1.1",
    "GFDL-1.1-only",
    "GFDL-1.2",
    "GFDL-1.2-only",
    "GFDL-1.2-or-later",
    "GFDL-1.3-no-cover-texts-no-invariant-sections",
    "GL2PS",
    "GPL-1.0+",
    "GPL-1.0-or-later",
    "GPL-2.0",
    "GPL-2.0+",
    "GPL-2.0-with-GCC-exception",
    "GPL-2.0-with-bison-exception",
    "GPL-2.0-with-classpath-exception",
    "GPL-3.0",
    "GPL-3.0-only",
    "GPL-3.0-or-later",
    "GPL-3.0-with-GCC-exception",
    "Giftware",
    "Glulxe",
    "HPND",
    "HaskellReport",
    "IBM-pibs",
    "ICU",
    "IJG",
    "IPA",
    "IPL-1.0",
    "ISC",
    "ImageMagick",
    "Imlib2",
    "Intel",
    "Intel-ACPI",
    "JSON",
    "LGPL-2.0",
    "LGPL-2.0-or-later",
    "LGPL-2.1",
    "LGPL-2.1-only",
    "LGPL-3.0",
    "LGPL-3.0-or-later",
    "LGPLLR",
    "LPL-1.0",
    "LPL-1.02",
    "LPPL-1.0",
    "LPPL-1.2",
    "LPPL-1.3c",
    "LiLiQ-R-1.1",
    "LiLiQ-Rplus-1.1",
    "Linux-OpenIB",
    "MIT",
    "MIT-advertising",
    "MIT-enna",
    "MPL-1.0",
    "MPL-1.1",
    "MPL-2.0",
    "MPL-2.0-no-copyleft-exception",
    "MS-PL",
    "MS-RL",
    "MirOS",
    "Motosoto",
    "Multics",
    "Mup",
    "NASA-1.3",
    "NCSA",
    "NGPL",
    "NOSL",
    "NPL-1.1",
    "NPOSL-3.0",
    "NTP",
    "Naumen",
    "Newsletr",
    "Nokia",
    "Noweb",
    "Nunit",
    "OCCT-PL",
    "OCLC-2.0",
    "ODC-By-1.0",
    "ODC-PDDL-1.0",
    "ODbL-1.0",
    "OFL-1.0",
    "OFL-1.1",
    "OGL-Canada-2.0",
    "OGL-UK-1.0",
    "OGL-UK-2.0",
    "OGL-UK-3.0",
    "OGTSL",
    "OLDAP-1.2",
    "OLDAP-1.3",
    "OLDAP-1.4",
    "OLDAP-2.0",
    "OLDAP-2.0.1",
    "OLDAP-2.1",
    "OLDAP-2.2",
    "OLDAP-2.2.1",
    "OLDAP-2.2.2",
    "OLDAP-2.3",
    "OLDAP-2.4",
    "OLDAP-2.6",
    "OLDAP-2.8",
    "OSET-PL-2.1",
    "OSL-1.0",
    "OSL-1.1",
    "OSL-2.0",
    "OSL-2.1",
    "OSL-3.0",
    "OpenSSL",
    "PHP-3.0",
    "PHP-3.01",
    "Plexus",
    "PostgreSQL",
    "Python-2.0",
    "QPL-1.0",
    "Qhull",
    "RHeCos-1.1",
    "RPL-1.1",
    "RPL-1.5",
    "RPSL-1.0",
    "RSA-MD",
    "RSCPL",
    "Ruby",
    "SAX-PD",
    "SCEA",
    "SGI-B-2.0",
    "SISSL",
    "SMLNJ",
    "SPL-1.0",
    "SWL",
    "Sendmail",
    "Sendmail-8.23",
    "SimPL-2.0",
    "Sleepycat",
    "Spencer-94",
    "Spencer-99",
    "SugarCRM-1.1.3",
    "TCL",
    "TCP-wrappers",
    "TOSL",
    "TU-Berlin-2.0",
    "Unicode-DFS-2015",
    "Unicode-TOU",
    "Unlicense",
    "VSL-1.0",
    "Vim",
    "W3C",
    "W3C-20150513",
    "Watcom-1.0",
    "X11",
    "XFree86-1.1",
    "XSkat",
    "Xerox",
    "Xnet",
    "ZPL-1.1",
    "ZPL-2.0",
    "Zed",
    "Zend-2.0",
    "Zimbra-1.3",
    "Zlib",
    "bsd-license",
    "bzip2-1.0.5",
    "canada-crown",
    "cc-nc",
    "curl",
    "diffmark",
    "dli-model-use",
    "dvipdfm",
    "eCos-2.0",
    "eGenix",
    "eurofound",
    "geo-no-fee-unrestricted",
    "geogratis",
    "gnuplot",
    "hesa-withrights",
    "jabber-osl",
    "libtiff",
    "localauth-withrights",
    "lucent-plan9",
    "met-office-cp",
    "mitre",
    "mpich2",
    "notspecified",
    "other-at",
    "other-closed",
    "other-nc",
    "other-open",
    "other-pd",
    "psfrag",
    "psutils",
    "ukclickusepsi",
    "ukcrown",
    "ukcrown-withrights",
    "ukpsi",
    "user-jsim",
    "wxWindows",
    "xpp",
    "zlib-acknowledgement",
}


def check_standard_title(title):
    """
    Assert a title conforms to the standard format.

    > CLDF dataset derived from AUTHOR's "TITLE" from YEAR

    Usage: In a dataset's `test.py`:
    ```python
    import os, pytest

    @pytest.mark.skipif("TRAVIS" in os.environ and os.environ["TRAVIS"] == "true")
    def test_valid_title(cldf_dataset, cldf_logger):
        from pylexibank import check_standard_title
        check_standard_title(cldf_dataset.metadata_dict['dc:title'])
    ```
    """
    match = STANDARD_TITLE_PATTERN.fullmatch(title)
    assert match and match.group('authors').strip().endswith(("'s", "s'"))


def get_creators_and_contributors(fname, strict=True):
    ctypes = {c.lower(): c for c in CONTRIBUTOR_TYPES}
    creators, contributors = [], []
    for row in iter_rows(fname):
        row = {k.lower(): v for k, v in row.items()}
        for role in nfilter([r.strip().lower() for r in row.get('role', '').split(',')]):
            c = {k: v for k, v in row.items() if k != 'role'}
            if role in {'author', 'creator', 'maintainer'}:
                creators.append(c)
            else:
                if strict:
                    c['type'] = ctypes[role]
                else:
                    c['type'] = ctypes.get(role, 'Other')
                contributors.append(c)
    return creators, contributors


def iter_rows(fname_or_lines):
    header, in_table = None, False

    def row(line):
        return [li.strip() for li in line.split('|')]

    for line in (
        fname_or_lines if isinstance(fname_or_lines, list)
        else fname_or_lines.open(encoding='utf-8')
    ):
        if in_table:
            if '|' not in line:  # Last row of table was already read
                break
            yield collections.OrderedDict(zip(header, row(line)))
        elif '|' in line:
            r = row(line)
            if not ''.join(r).replace('-', ''):  # a line like " --- | --- | --- "
                assert header
                in_table = True
            else:
                header = r


@attr.s
class LexibankMetadata(Metadata):
    aboutUrl = attr.ib(default=None)
    conceptlist = attr.ib(
        default=[],
        converter=lambda s: [] if not s else (s if isinstance(s, list) else [s]))
    lingpy_schema = attr.ib(default=None)
    derived_from = attr.ib(default=None)
    related = attr.ib(default=None)
    source = attr.ib(default=None)
    patron = attr.ib(default=None)
    version = attr.ib(default=None)

    def __attrs_post_init__(self):
        import pylexibank

        self.version = self.version or pylexibank.__version__

    @property
    def zenodo_license(self):
        if self.known_license and self.known_license.id in LICENSES:
            return self.known_license.id

    def common_props(self):
        res = super().common_props()
        res.update({
            "dc:format": [
                "http://concepticon.clld.org/contributions/{0}".format(cl)
                for cl in self.conceptlist],
            "dc:isVersionOf": "http://lexibank.clld.org/contributions/{0}".format(
                self.derived_from) if self.derived_from else None,
            "dc:related": self.related,
            "aboutUrl": self.aboutUrl
        })
        return res

    def markdown(self):
        lines = [super().markdown(), '']

        if self.related:
            lines.append('See also %s\n' % self.related)

        if self.conceptlist:
            lines.append('Conceptlists in Concepticon:')
            lines.extend([
                '- [{0}](https://concepticon.clld.org/contributions/{0})'.format(cl)
                for cl in self.conceptlist])
            lines.append('')
        return '\n'.join(lines)



"""
Check for (potentially) problematic consonant clusters >= 3.
"""

import collections

from cldfbench.cli_util import with_dataset, add_catalog_spec
from clldutils.clilib import Table, add_format

from pylexibank.cli_util import add_dataset_spec


def register(parser):
    add_dataset_spec(parser)
    add_catalog_spec(parser, "clts")
    add_format(parser, default="pipe")

    parser.add_argument(
        "-l", "--length",
        type=int,
        default=3,
        help="Check for consonant clusters of this length or more"
    )


def run(args):
    with_dataset(args, collect_consonant_clusters)


def compute_consonant_cluster(word, sound_names):
    out = [[]] if sound_names[0].split(" ")[-1] in ["consonant", "cluster"] else []
    for i, sound in enumerate(sound_names):
        if sound.split(" ")[-1] in ["diphthong", "vowel", "tone", "�", "marker"]:
            out += [[]]
        else:
            out[-1] += [word[i]]
    return [chunk for chunk in out if chunk]


def collect_consonant_clusters(dataset, args):
    by_lang = collections.defaultdict(lambda: collections.defaultdict(list))

    with Table(args, "Language_ID", "Length", "Cluster", "Words") as table:
        for row in sorted(
            dataset.cldf_dir.read_csv("forms.csv", dicts=True), key=lambda r: r["ID"]
        ):
            if "<<" in row["Segments"]:
                args.log.warning(
                    "Invalid segments in {0} (ID: {1}).".format(row["Segments"], row["ID"]))
                continue
            else:
                segments = row["Segments"].split(" + ")

            for morpheme, sounds in map(
                lambda x: (x.split(), [s.name for s in args.clts.api.bipa(x.split())]),
                segments
            ):
                clusters = compute_consonant_cluster(morpheme, sounds)

                for cluster in clusters:
                    by_lang[tuple(cluster)][row["Language_ID"]] += [row["Segments"], row["Form"]]

        cases = 0
        for cluster in sorted(by_lang, key=lambda x: len(x)):
            data = by_lang[cluster]
            if len(cluster) >= args.length:
                cases += 1
                for language, words in data.items():
                    table.append(
                        [language, str(len(cluster)), " ".join(cluster), " // ".join(words)])

    args.log.warning(
        f"Found {cases} potentially problematic consonant cluster(s) with length {args.length}.")



from collections import Counter
from pathlib import Path

import pytest
from csvw.dsv import reader

from pylexibank import util


@pytest.mark.parametrize(
    'seq,subseq,repl,out',
    [
        ('abcdabcd', 'bc', 'x', list('axdaxd')),
        ([1, 2, 3, 4], [2], [9, 9], [1, 9, 9, 3, 4]),
    ]
)
def test_iter_repl(seq, subseq, repl, out):
    assert list(util.iter_repl(seq, subseq, repl)) == out


def test_jsondump(tmp_path):
    fname = tmp_path / 'dump.json'
    res = util.jsondump({'a': 2}, fname)
    assert 'a' in res
    res = util.jsondump({'b': 3}, fname)
    assert res['b'] == 3 and res['a'] == 2


def test_getEvoBibAsBibtex(mocker):
    bib = '<pre>@book{key,\ntitle={The Title}\n}\n</pre>'
    mocker.patch(
        'pylexibank.util.get_url', mocker.Mock(return_value=mocker.Mock(text=bib)))
    assert '@book' in util.getEvoBibAsBibtex('')


class MockResponse(object):
    def __init__(self, p):
        p = Path(p)
        self.status_code = 200 if p.exists() else 404
        self.path = p

    def iter_content(self, *args, **kw):
        if self.path.exists():
            with open(self.path.as_posix(), 'rb') as fp:
                yield fp.read()


class MockRequests(object):
    def __init__(self, p):
        self.path = p

    def get(self, *args, **kw):
        return MockResponse(self.path)


def test_split_by_year():
    assert util.split_by_year(' 2012. abc') == ('', '2012', 'abc')
    assert util.split_by_year(' (2012) abc') == ('', '2012', 'abc')
    assert util.split_by_year('abc') == (None, None, 'abc')


def test_get_badge():
    for r in util.pb(list(range(10))):
        util.get_badge((r / 10.0) + 0.5, 'name')


def test_get_reference():
    assert util.get_reference('John Doe', '1998', 'The Title', None, {})
    assert util.get_reference(None, None, None, None, {}) is None
    assert util.get_reference(None, None, 'The Title', None, {}).source.id == 'thetitle'


def test_sorted_obj():
    d1 = {'a': [1, 2, 3], 'b': dict(a=3, b=1)}
    d2 = {'b': Counter('baaa'), 'a': [1, 2, 3]}
    assert util.sorted_obj(d1) == util.sorted_obj(d2)
    assert util.sorted_obj(d2)['b']['a'] == 3
    util.sorted_obj(['http://www.w3.org/ns/csvw', {'@language': 'en'}])


def test_get_concepts(concepticon):
    res = util.get_concepts(concepticon.conceptlists.values(), [])
    assert len(res) == 1
    assert 'chinese' in res[0].attributes
    assert res[0].number == '1'

    id_lookup, _ = util.get_ids_and_attrs(
        res, {}, id_factory=lambda c: c.number + 'x', lookup_factory=lambda c: c.number + 'y')
    assert id_lookup['1y'] == '1x'

    id_lookup, _ = util.get_ids_and_attrs(
        res,
        {'number': 'Number'},
        id_factory=lambda c: c.number + 'x',
        lookup_factory=lambda c: c['Number'] + 'y')
    assert id_lookup['1y'] == '1x'

    csv = Path(__file__).parent / 'repos' / 'datasets' / 'test_dataset' / 'etc' / 'concepts.csv'
    res = util.get_concepts([], list(reader(csv, dicts=True)))
    assert len(res) == 2
    assert 'chinese' in res[0].attributes
    assert res[0].number == '1'

    id_lookup, _ = util.get_ids_and_attrs(
        res,
        {'chinese': 'chi'},
        id_factory=lambda c: c.number + 'x',
        lookup_factory=lambda c: c['chi'])
    assert id_lookup['xyz'] == '1x'



import pytest

from pylexibank.forms import *


def test_split():
    spec = FormSpec()
    assert spec.split({}, 'x', lexemes={'x': 'x(a);y'}) == ['x', 'y']
    assert spec.split({}, 'x', lexemes={'x': 'x;?'}) == ['x']

    spec.strip_inside_brackets = False
    assert spec.split({}, 'x(a)') == ['x(a)']

    spec = FormSpec(separators='|')
    assert spec.split(None, 'x|y') == ['x', 'y']
    assert spec.split(None, 'x;y') == ['x;y']

    spec = FormSpec(first_form_only=True)
    assert spec.split({}, 'x', lexemes={'x': 'x;y'}) == ['x']


def test_markdown():
    spec = FormSpec()
    assert spec.as_markdown()


def test_normalize_whitespace():
    spec = FormSpec()
    assert spec.clean(' a\t b\n') == 'a b'

    spec = FormSpec(normalize_whitespace=False, strip_inside_brackets=False)
    assert spec.clean(' a\t b\n') == ' a\t b\n'


def test_normalize_unicode():
    spec = FormSpec(normalize_unicode='NFD', separators='\u0308')
    # The combining diaresis is used as separator:
    assert len(spec.split(None, 'äb')) == 2


def test_separators():
    with pytest.raises(ValueError):
        FormSpec(separators=5)

    with pytest.raises(ValueError):
        FormSpec(separators=['ab'])


def test_replacements():
    spec = FormSpec(replacements=[('x', 'y')])
    assert spec.clean('x') == 'y'
    assert spec.clean('y') == 'y'

    with pytest.raises(ValueError):
        FormSpec(replacements=())

    with pytest.raises(ValueError):
        FormSpec(replacements=[(1, 2)])



import pytest

import pyclts

from pylexibank.transcription import analyze, Analysis


def test_analyze(repos):
    clts = pyclts.CLTS(repos)
    with pytest.raises(ValueError):
        analyze(clts, [], Analysis())

    with pytest.raises(ValueError):
        analyze(clts, ['\n'], Analysis())

    segments, la, clpa, analysis = analyze(clts, ['a', '^', 'b'], Analysis())
    assert segments == ['a', '^', 'b']
    assert analysis.general_errors == 1



import re
from textwrap import dedent

from bs4 import BeautifulSoup
import bs4
from csvw.dsv import UnicodeWriter
from clldutils.misc import slug

from pylexibank.dataset import Dataset
from pylexibank.forms import FormSpec
from pylexibank.util import pb


class TOB(Dataset):
    name = None
    dset = None
    pages = 1
    lexemes = {}

    form_spec = FormSpec(
        brackets={"(": ")"},
        separators=";/,~",
        missing_data=('?', '-', ''),
        strip_inside_brackets=True
    )

    tob_sources = {
        "Starostin2011": dedent("""
        @misc{Starostin2011,
            Year = {2011},
            Editor = {Starostin, George S. and Krylov, Phil},
            Title = {The Global Lexicostatistical Database}
        }""")
    }

    def _url(self, page):
        return 'http://starling.rinet.ru/cgi-bin/response.cgi?' + \
            'root=new100&morpho=0&basename=new100' + \
            r'\{0}\{1}&first={2}'.format(self.dset, self.name, page)

    def cmd_download(self, args):
        self.raw_dir.write('sources.bib', "\n".join(self.tob_sources.values()))

        # download data
        all_records = []
        for i in pb(list(range(1, 20 * self.pages + 1, 20))):
            with self.raw_dir.temp_download(
                    self._url(i), 'file-{0}'.format(i), log=args.log) as fname:
                soup = BeautifulSoup(fname.open(encoding='utf8').read(), 'html.parser')
                for record in soup.findAll(name='div', attrs={"class": "results_record"}):
                    if isinstance(record, bs4.element.Tag):
                        children = list(record.children)
                        number = children[0].findAll('span')[1].text.strip()
                        concept = children[1].findAll('span')[1].text
                        for child in children[2:]:
                            if isinstance(child, bs4.element.Tag):
                                dpoints = child.findAll('span')
                                if len(dpoints) >= 3:
                                    lname = dpoints[1].text
                                    glottolog = re.findall(
                                        'Glottolog: (........)', str(dpoints[1]))[0]
                                    entry = dpoints[2].text
                                    cogid = list(child.children)[4].text.strip()
                                    all_records.append(
                                        (number, concept, lname, glottolog, entry, cogid))
        with UnicodeWriter(self.raw_dir / 'output.csv') as f:
            f.writerows(all_records)

    def cmd_makecldf(self, args):
        args.writer.add_sources()
        concepts = args.writer.add_concepts(
            id_factory=lambda c: c.id.split('-')[-1] + '_' + slug(c.english),
            lookup_factory=lambda c: c.id.split('-')[-1]
        )
        for cid, concept, lang, gc, form, cogid in self.raw_dir.read_csv('output.csv'):
            lid = lang.replace(' ', '_')
            args.writer.add_language(ID=lid, Name=lang, Glottocode=gc)
            global_cog_id = "%s-%s" % (cid, cogid)
            for row in args.writer.add_forms_from_value(
                Language_ID=lid,
                Parameter_ID=concepts[cid],
                Value=form,
                Source=list(self.tob_sources),
                Cognacy=global_cog_id
            ):
                args.writer.add_cognate(
                    lexeme=row,
                    Cognateset_ID=global_cog_id,
                    Source=list(self.tob_sources))



import pathlib
import collections
import unicodedata

import attr

from csvw.dsv import reader
from clldutils.misc import lazyproperty
from clldutils import jsonlib
from pyglottolog.languoids import Glottocode

from segments import Tokenizer

from cldfbench.dataset import Dataset as BaseDataset
from cldfbench.cldf import CLDFSpec

from pylexibank.util import jsondump, log_dump
from pylexibank import cldf
from pylexibank import models
from pylexibank import transcription
from pylexibank import metadata
from pylexibank import forms
from pylexibank import report
from pylexibank.profile import Profile
from pylexibank.lingpy_util import settings

__all__ = ['Dataset', 'ENTRY_POINT']
ENTRY_POINT = 'lexibank.dataset'


class Dataset(BaseDataset):
    """
    A lexibank dataset.

    This object provides access to a dataset's
    - language list as attribute `languages`
    - concept list as attribute `concepts`
    - concepticon concept-list ID as attribute `conceptlist`
    """
    metadata_cls = metadata.LexibankMetadata
    # The CLDFWriter can be instructed to keep only objects which are referenced from FormTable by
    # setting the following options to `False`.
    writer_options = dict(keep_languages=True, keep_parameters=True)

    lexeme_class = models.Lexeme
    cognate_class = models.Cognate
    language_class = models.Language
    concept_class = models.Concept

    form_spec = forms.FormSpec()

    # If a dataset provides cross-concept cognate sets, it must declare this by setting the below
    # flag to True.
    cross_concept_cognates = False

    @property
    def stats(self):
        if self.dir.joinpath('README.json').exists():
            return jsonlib.load(self.dir / 'README.json')
        return {}

    def __init__(self, concepticon=None, glottolog=None):
        super().__init__()
        if self.__class__ != Dataset:
            if not self.id:
                raise ValueError(
                    "Dataset.id needs to be specified in subclass for %s!" % self.__class__)
        self.unmapped = Unmapped()
        self._json = self.dir / 'lexibank.json'
        self.contributors_path = self.dir / 'CONTRIBUTORS.md'

        self.conceptlists = []
        self.glottolog = glottolog
        self.concepticon = concepticon
        self.tr_analyses = {}
        self.tr_bad_words = []
        self.tr_invalid_words = []

    def cldf_specs(self):
        return CLDFSpec(
            module='Wordlist',
            writer_cls=cldf.LexibankWriter,
            dir=self.cldf_dir,
            metadata_fname=cldf.MD_NAME,
            default_metadata_path=pathlib.Path(__file__).parent / cldf.MD_NAME)

    def _iter_etc(self, what):
        delimiter = '\t'
        path = self.etc_dir / (what + '.tsv')
        if not path.exists():
            delimiter = ','
            path = path.parent / (what + '.csv')
        return reader(path, dicts=True, delimiter=delimiter) if path.exists() else []

    def read_json(self):  # pragma: no cover
        return jsonlib.load(self._json) if self._json.exists() else {}

    def write_json(self, obj):  # pragma: no cover
        jsondump(obj, self._json)

    def get_creators_and_contributors(self, strict=False):
        if self.contributors_path.exists():
            return metadata.get_creators_and_contributors(self.contributors_path, strict=strict)
        return [], []

    @lazyproperty
    def sources(self):
        return list(self._iter_etc('sources'))

    @lazyproperty
    def concepts(self):
        return list(self._iter_etc('concepts'))

    @lazyproperty
    def languages(self):
        res = []
        for item in self._iter_etc('languages'):
            if item.get('GLOTTOCODE', None) and not \
                    Glottocode.pattern.match(item['GLOTTOCODE']):  # pragma: no cover
                raise ValueError(
                    "Invalid glottocode {0}".format(item['GLOTTOCODE']))
            res.append(item)
        return res

    def _replacements(self, what, source_col, target_col='REPLACEMENT'):
        return collections.OrderedDict(
            [(item[source_col], item[target_col]) for item in self._iter_etc(what)])

    @lazyproperty
    def lexemes(self):
        return self._replacements('lexemes', 'LEXEME')

    @lazyproperty
    def segments(self):
        return self._replacements('segments', 'SEGMENT')

    # ---------------------------------------------------------------
    # handling of lexemes/forms/words
    # ---------------------------------------------------------------
    @lazyproperty
    def orthography_profile_dict(self):
        res = {}
        profile = self.etc_dir / 'orthography.tsv'
        profile_dir = self.etc_dir / 'orthography'
        if profile.exists():
            res[None] = profile
        if profile_dir.exists() and profile_dir.is_dir():
            for p in profile_dir.glob('*.tsv'):
                res[p.stem] = p

        return {k: Profile.from_file(str(p), form='NFC') for k, p in res.items()}

    @staticmethod
    def form_for_segmentation(form):
        return unicodedata.normalize('NFC', '^' + form + '$')

    @lazyproperty
    def tokenizer(self):
        """
        Datasets can provide support for segmentation (aka tokenization) in two ways:
        - by providing an orthography profile at etc/orthography.tsv or
        - by overwriting this method to return a custom tokenizer callable.

        :return: A callable to do segmentation.

        The expected signature of the callable is

            def t(item, string, **kw)

        where
        - `item` is a `dict` representing the complete CLDF FormTable row
        - `string` is the string to be segmented
        - `kw` may be used to pass any context info to the tokenizer, when called
          explicitly.
        """
        tokenizers = {
            k: Tokenizer(profile=p, errors_replace=lambda c: '<{0}>'.format(c))
            for k, p in self.orthography_profile_dict.items()}

        if tokenizers:
            def _tokenizer(item, string, **kw):
                """
                Adds `Profile` and `Graphemes` keys to `item`, returns `list` of segments.
                """
                kw.setdefault("column", "IPA")
                kw.setdefault("separator", " + ")
                profile = kw.pop('profile', None)
                if profile:
                    tokenizer = tokenizers[profile]
                    item['Profile'] = profile
                elif isinstance(item, dict) \
                        and 'Language_ID' in item \
                        and item['Language_ID'] in tokenizers:
                    tokenizer = tokenizers[item['Language_ID']]
                    item['Profile'] = item['Language_ID']
                else:
                    tokenizer = tokenizers[None]
                    item['Profile'] = 'default'
                form = self.form_for_segmentation(string)
                res = tokenizer(form, **kw).split()
                kw['column'] = Profile.GRAPHEME_COL
                item['Graphemes'] = tokenizer(form, **kw)
                return res
            return _tokenizer

    def _cmd_makecldf(self, args):
        # Inject the appropriate CLDFWriter instance:
        self.unmapped.clear()
        self.tr_analyses = {}
        self.tr_bad_words = []
        self.tr_invalid_words = []

        if len(self.metadata.conceptlist):
            self.conceptlists = [
                self.concepticon.conceptlists[key] for key in self.metadata.conceptlist]

        if self.concept_class is models.CONCEPTICON_CONCEPTS:
            assert self.conceptlists
            self.concept_class = models.concepticon_concepts(self.conceptlists)

        super()._cmd_makecldf(args)

        #
        # make sure properties have the appropriate datatypes:
        #
        ds = self.cldf_reader()
        for col in ds['LanguageTable'].tableSchema.columns:
            if col.name.lower() == 'population':  # pragma: no cover
                assert col.datatype.base == 'integer', 'population must be integer!'

        if args.verbose:
            self.unmapped.pprint()
        if not args.dev:
            assert self.cldf_reader().validate(args.log)

        stats = transcription.Stats(
            bad_words=sorted(self.tr_bad_words[:100], key=lambda x: x['ID']),
            bad_words_count=len(self.tr_bad_words),
            invalid_words=sorted(self.tr_invalid_words[:100], key=lambda x: x['ID']),
            invalid_words_count=len(self.tr_invalid_words))
        for lid, analysis in self.tr_analyses.items():
            for attribute in ['segments', 'bipa_errors', 'sclass_errors', 'replacements']:
                getattr(stats, attribute).update(getattr(analysis, attribute))
            stats.general_errors += analysis.general_errors
            stats.inventory_size += len(analysis.segments) / len(self.tr_analyses)

        error_segments = stats.bipa_errors.union(stats.sclass_errors)
        for i, row in enumerate(stats.bad_words):
            analyzed_segments = []
            for s in row['Segments']:
                analyzed_segments.append('<s> %s </s>' % s if s in error_segments else s)
            stats.bad_words[i] = [
                row['ID'],
                row['Language_ID'],
                row['Parameter_ID'],
                row['Form'],
                ' '.join(analyzed_segments)]

        for i, row in enumerate(stats.invalid_words):
            stats.invalid_words[i] = [
                row['ID'],
                row['Language_ID'],
                row['Parameter_ID'],
                row['Form']]
        # Aggregate transcription analysis results ...
        tr = dict(
            by_language={k: attr.asdict(v) for k, v in self.tr_analyses.items()},
            stats=attr.asdict(stats))

        jsondump(tr, self.cldf_dir / '.transcription-report.json', log=args.log)

        # ... and write a report:
        args.tr_analysis = tr
        self._cmd_readme(args)
        (self.dir / 'TRANSCRIPTION.md').write_text(transcription.report(tr), encoding='utf8')
        log_dump(self.dir / 'TRANSCRIPTION.md', args.log)

        jsondump(settings(), self.cldf_dir / 'lingpy-rcParams.json', log=args.log)

    def cmd_readme(self, args):
        res = self.metadata.markdown()
        tr = self.cldf_dir / '.transcription-report.json'
        tr = jsonlib.load(tr) if tr.exists() else None
        res += report.report(
            self,
            tr,
            None if args.dev else getattr(args, 'glottolog', None),
            args.log,
        )
        if self.contributors_path.exists():
            res += '\n\n{0}\n\n'.format(self.contributors_path.read_text(encoding='utf8'))
        self.dir.write('FORMS.md', self.form_spec.as_markdown(self))
        return res


class Unmapped(object):
    def __init__(self):
        self.languages = set()
        self.concepts = set()

    def clear(self):
        self.languages = set()
        self.concepts = set()

    def add_concept(self, **kw):
        self.concepts.add(models.Concept(**kw))

    def add_language(self, **kw):
        self.languages.add(models.Language(**kw))

    @staticmethod
    def quote(v):
        v = '{0}'.format(v or '')
        if ',' in v or len(v.split()) > 1:
            v = '"%s"' % v.replace('"', '""')
        return v

    def pprint(self):
        for objs, cls in [(self.languages, models.Language), (self.concepts, models.Concept)]:
            if objs:
                print('=== Unmapped %ss ===' % cls.__name__)
                print(','.join([a.name.upper() for a in attr.fields(cls)]))
                for row in sorted(map(attr.astuple, objs)):
                    print(','.join(map(self.quote, row)))



"""
Orthography profiles in lexibank are based on the Profile class in the `segments` package, adding
functionality to check and consistently format in a diff-friendly way.
"""
import re
import copy
import collections

import pyclts
import segments
from segments.tree import Tree
from clldutils.misc import log_or_raise
from csvw import dsv
from csvw.metadata import TableGroup, Column

__all__ = ['Profile', 'IPA_COLUMN']

IPA_COLUMN = 'IPA'


def unicode2codepointstr(text):
    return " ".join(["U+{0:0{1}X}".format(ord(char), 4) for char in text])


def ipa2tokens(text):
    return [t if "/" not in t else t.split("/")[1] for t in (text or '').split()]


def ipa2sca(text, clts):
    sca = clts.soundclass("sca")
    return " ".join(
        clts.bipa.translate(t, sca) if t != "NULL" else "NULL" for t in ipa2tokens(text))


class Profile(segments.Profile):
    def __init__(self, *specs, **kw):
        super().__init__(*specs, **kw)
        default_spec = list(next(iter(self.graphemes.values())).keys())
        for grapheme in ['^', '$']:
            if grapheme not in self.graphemes:
                self.graphemes[grapheme] = {k: None for k in default_spec}
        self.recreate_tree()

    def __str__(self):
        # We overwrite the base class' method to fix the order of columns.
        tg = TableGroup.fromvalue(self.MD)
        for col in sorted(
                self.column_labels, key=lambda t: (t == IPA_COLUMN, t.lower()), reverse=True):
            if col != self.GRAPHEME_COL:
                tg.tables[0].tableSchema.columns.append(
                    Column.fromvalue({"name": col, "null": self.NULL}))

        return tg.tables[0].write(self.iteritems(), fname=None).decode('utf8').strip()

    def write(self, fname=None):
        (fname or self.fname).write_text(str(self), encoding='utf8')

    def recreate_tree(self):
        self.tree = Tree(list(self.graphemes.keys()))

    def sort(self, clts=None, ipa_col=IPA_COLUMN):
        self.graphemes = collections.OrderedDict(
            sorted(
                self.graphemes.items(),
                key=lambda e: (
                    e[0] not in ["^", "$"],
                    e[0] != "^",
                    e[1][ipa_col] != None,
                    re.match(r"\^.*\$", e[0]) is None,
                    len(ipa2sca(e[1][ipa_col], clts)) if clts else False,
                    ipa2sca(e[1][ipa_col], clts) if clts else False,
                    len(e[0]),
                    e[0],
                ),
            )
        )

    def trim(self, ipa_col=IPA_COLUMN):
        # Make a copy of the profile (so we don't change in place)
        new_profile = collections.OrderedDict()
        for g, entry in self.graphemes.items():
            spec = copy.copy(entry)
            spec[self.GRAPHEME_COL] = g
            new_profile[g] = spec

        # Collect all keys, so that we will gradually remove them; those with
        # ^ and $ go first
        graphemes = list(new_profile.keys())
        bound_graphemes = [
            grapheme for grapheme in graphemes if grapheme[0] == "^" and grapheme[-1] == "$"]
        bound_graphemes += [
            grapheme for grapheme in graphemes if grapheme[0] == "^" and grapheme[-1] != "$"]
        bound_graphemes += [
            grapheme for grapheme in graphemes if grapheme[0] != "^" and grapheme[-1] == "$"]

        check_graphemes = bound_graphemes + sorted(
            [
                grapheme
                for grapheme in graphemes if len(grapheme) > 1 and grapheme not in bound_graphemes
            ],
            key=len,
            reverse=True,
        )

        # For each entry, we will remove it from `segment_map`, apply the resulting
        # profile, and add the entry back at the end of loop (still expansive, but
        # orders of magnitude less expansive than making a copy at each iteration)
        removed = 0
        for grapheme in check_graphemes:
            if grapheme in new_profile:
                ipa = new_profile[grapheme][ipa_col]
                # Obtain the segments without the current rule
                t = segments.Tokenizer(
                    profile=Profile(
                        *[copy.copy(s) for g, s in new_profile.items() if g != grapheme]))
                if t(grapheme, column=ipa_col) == ipa:
                    # If the resulting `segments` match the `ipa` reference, we can delete the rule:
                    removed += 1
                    del new_profile[grapheme]

        for g in set(self.graphemes.keys()) - set(new_profile.keys()):
            del self.graphemes[g]

        self.recreate_tree()
        return removed

    @staticmethod
    def segmentable_form(form):
        form = form.strip()
        if not form.startswith('^'):
            form = '^' + form
        if not form.endswith('$'):
            form += '$'
        return form

    def augment(self, forms, clts=None, ipa_col=IPA_COLUMN):
        """
        Applies a profile to a wordlist, returning new profile counts and segments.
        """
        self.column_labels.add('FREQUENCY')
        if clts:
            self.column_labels.add('SCA')
        self.column_labels.add('EXAMPLES')
        freqs = collections.Counter()
        ex = collections.defaultdict(list)
        t = segments.Tokenizer(profile=self)
        for form in forms:
            graphemes = t(self.segmentable_form(form)).split()
            freqs.update(graphemes)
            for g in graphemes:
                ex[g].append(form[1:-1])
        for g, spec in self.graphemes.items():
            spec['FREQUENCY'] = freqs.get(g, 0)
            spec['EXAMPLES'] = ";".join(ex.get(g, [])[:5])
            if clts:
                spec['SCA'] = ipa2sca(spec[ipa_col], clts)

    def clean(self, clts, ipa_col=IPA_COLUMN):
        """
        Replace user-provided IPA graphemes with the CLTS/BIPA default ones.
        """
        def clean_segment(segment, clts):
            if "/" in segment:
                left, right = segment.split("/")
                return "%s/%s" % (left, str(clts.bipa[right]))
            return str(clts.bipa[segment])

        for grapheme, entry in self.graphemes.items():
            # Remove any multiple spaces, split IPA first into segments and then
            # left- and right- slash information (if any), and use the default
            if entry[ipa_col]:
                ipa_value = re.sub(r"\s+", " ", entry[ipa_col]).strip()
                entry[ipa_col] = " ".join(
                    clean_segment(segment, clts) for segment in ipa_value.split())
                if 'CODEPOINTS' in self.column_labels:
                    entry["CODEPOINTS"] = unicode2codepointstr(grapheme)

    def check(self, clts=None, log=None, ipa_col=IPA_COLUMN):
        """
        Check a profile for consistency, logging problems.

        For each grapheme, raise:
        - a warning if there are duplicate entries
        - an error if there are inconsistencies
        - an error if the mapping has invalid BIPA
        """
        mapping = collections.defaultdict(list)
        if self.fname:
            # We read the profile from disk because segments.Profile already skips duplicate
            # graphemes, which we want to investigate more closely.
            for spec in dsv.reader(self.fname, dicts=True, delimiter='\t'):
                mapping[spec[self.GRAPHEME_COL]].append(spec[ipa_col])

        for grapheme in mapping:
            # check mapping consistency
            if len(mapping[grapheme]) >= 2:
                if len(set(mapping[grapheme])) == 1:
                    log_or_raise(
                        "Duplicate, redundant entry or entries for grapheme [{}].".format(grapheme),
                        log=log,
                        level='warning')
                else:
                    log_or_raise(
                        "Inconsist entries for grapheme [{}]: multiple mappings {}.".format(
                            grapheme,
                            str(mapping[grapheme])),
                        log=log,
                        level='error')

            # check BIPA consistency
            if clts:
                for value in mapping[grapheme]:
                    if value:
                        # check for unknown sounds
                        unknown = [
                            isinstance(clts.bipa[segment], pyclts.models.UnknownSound)
                            for segment in ipa2tokens(value) if segment and segment != 'NULL']
                        if any(unknown):
                            log_or_raise(
                                "Mapping [{}] ({}) -> [{}] ({}) includes an unknown sound.".format(
                                    grapheme,
                                    unicode2codepointstr(grapheme),
                                    value,
                                    unicode2codepointstr(value)),
                                log=log,
                                level='error'
                            )



import sys
import json
import importlib

import pytest
from clldutils.path import sys_path
from cldfbench.dataset import NOOP

from pylexibank import Lexeme, Dataset
from pylexibank.dataset import Unmapped


def import_module(p):
    with sys_path(p.parent):
        if p.stem in sys.modules:
            return importlib.reload(sys.modules[p.stem])
        return importlib.import_module(p.stem)


def test_Item():
    with pytest.raises(TypeError):
        Lexeme()

    with pytest.raises(ValueError):
        Lexeme(ID=1, Form='a', Value='', Language_ID='l', Parameter_ID='p')


def test_Unmapped(capsys):
    unmapped = Unmapped()
    unmapped.add_language(ID='tl', Name='the language')
    unmapped.add_concept(ID='tc', Name='the concept')
    unmapped.pprint()
    out, err = capsys.readouterr()
    assert 'tc,"the concept",,' in out.split('\n')


def test_invalid_dataset():
    class Test(Dataset):
        pass

    with pytest.raises(ValueError):
        Test()


def test_Dataset_tokenizer(tmp_path):
    etc = tmp_path / 'etc'
    etc.mkdir()
    orth_dir = etc.joinpath('orthography')
    orth_dir.mkdir()
    orth_dir.joinpath('l1.tsv').write_text('Grapheme\tIPA\na\tb')

    class DS(Dataset):
        id = '1'
        dir = etc.parent

    ds = DS()
    assert ds.tokenizer({}, 'a', profile='l1') == ['b']
    assert ds.tokenizer({'Language_ID': 'l1'}, 'a') == ['b']

    etc.joinpath('orthography.tsv').write_text('Grapheme\tIPA\na\tc')
    ds = DS()
    assert ds.tokenizer({}, 'a') == ['c']


def test_BaseDataset(mocker, repos):
    class TestDataset(Dataset):
        dir = repos / 'datasets' / 'test_dataset'
        id = 'abc'

    ds = TestDataset(glottolog=mocker.Mock(), concepticon=mocker.Mock())
    assert ds.cmd_download(mocker.Mock()) == NOOP
    assert ds.cmd_makecldf(mocker.Mock()) == NOOP
    assert ds.sources
    assert ds.concepts
    assert ds.languages
    assert len(ds.raw_dir.read_bib('sources_ext.bib')) == 96

    assert not ds.stats
    ds.dir.write('README.json', json.dumps({'a': 1}))
    assert ds.stats['a'] == 1


@pytest.mark.parametrize(
    'string,segments,graphemes',
    [
        ('^b$', 'b', '^ ^ b $ $'),  # context marker is stripped
        ('aba', 'c b z', '^a b a$'),  # "a" is treated differently depending on context
        ('bab', 'b a b', '^ b a b $'),  # "a" is treated differently depending on context
    ]
)
def test_tokenizer(repos, string, segments, graphemes):
    import warnings

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        mod = import_module(repos / 'datasets' / 'test_dataset_cldf' / 'tdc')
        dataset = mod.Test()
        item = {}
        assert dataset.tokenizer(item, string) == segments.split()
        assert item['Graphemes'] == graphemes


def test_contributors(dataset, dataset_cldf):
    assert len(dataset.get_creators_and_contributors()[0]) == 1
    assert len(dataset_cldf.get_creators_and_contributors()[0]) == 0



"""
Format (and lint) the orthography profiles of a dataset.
"""
import collections

from cldfbench.cli_util import add_catalog_spec, get_dataset
from pylexibank.cli_util import add_dataset_spec
from pylexibank.profile import IPA_COLUMN


def register(parser):
    add_dataset_spec(parser)
    add_catalog_spec(parser, 'clts')

    parser.add_argument(
        "--ipa",
        help="Name of the profile column with the IPA representation.",
        default=IPA_COLUMN,
    )
    parser.add_argument(
        '--augment',
        help="augment the profile with sca, and frequency counts and examples from the FormTable",
        default=False,
        action='store_true',
    )
    parser.add_argument(
        '--trim',
        help="remove redundant orthography rules",
        default=False,
        action='store_true',
    )
    parser.add_argument(
        '--sort',
        help="sort the profile",
        default=False,
        action='store_true',
    )


def run(args):
    ds = get_dataset(args)
    clts = args.clts.api

    # Load the profile(s) specified for the dataset
    profiles = {k or 'default': v for k, v in ds.orthography_profile_dict.items()}
    forms = collections.defaultdict(list)
    if ds.cldf_dir.joinpath('forms.csv').exists():
        for form in ds.cldf_reader()['FormTable']:
            forms[form.get('Profile')].append(ds.form_for_segmentation(form['Form']))
    if list(forms.keys()) == [None]:  # pragma: no cover
        forms['default'] = forms[None]

    for key, profile in profiles.items():
        args.log.info('Processing {0}'.format(profile.fname))
        profile.clean(clts, ipa_col=args.ipa)

        if args.trim:
            # Run the trimmer as many times as necessary until nothing more is left to remove
            total_removed = 0
            while True:
                removed = profile.trim(ipa_col=args.ipa)
                total_removed += removed
                if removed == 0:
                    break
            if total_removed:  # pragma: no cover
                args.log.info("{} superfluous rules were removed.".format(total_removed))

        if args.augment and forms[key]:
            profile.augment(forms[key], clts=args.clts.api)

        if args.sort:
            profile.sort(clts=args.clts.api, ipa_col=args.ipa)

        profile.check(clts, args.log, ipa_col=args.ipa)
        profile.write()



import pytest
from csvw.metadata import Column

from pylexibank.db import Database, ColSpec, schema


@pytest.fixture
def db(tmp_path):
    return Database(tmp_path / 'lexibank.sqlite')


def test_ColSpec():
    col = ColSpec(name='c', csvw_type='float')
    assert col.convert(5) == '5'


def test_schema(cldf_dataset):
    cldf_dataset['ParameterTable', 'Name'].name = 'thename'
    assert cldf_dataset['ParameterTable', 'http://cldf.clld.org/v1.0/terms.rdf#name'].header == 'thename'
    tables, reftables = schema(cldf_dataset)
    assert len(tables) == 4
    assert len(reftables) == 2
    for t in tables:
        ptschema = t.sql
        if 'ParameterTable' in ptschema:
            assert "`Name`" in ptschema
            break
    else:
        assert False


def test_db(dataset, mocker, capsys, db):
    db.load(dataset)
    db.create(exists_ok=True)
    with pytest.raises(ValueError):
        db.create()
    db.create(force=True)
    db.load(dataset)
    db.load_glottolog_data(dataset.glottolog)
    db.load_concepticon_data(mocker.Mock(conceptsets={}))
    for sql in db.sql:
        db.fetchall(sql)
    with db.connection() as conn:
        db.fetchall('select * from dataset', conn=conn, verbose=True)
    out, _ = capsys.readouterr()
    assert 'select' in out

    db.create(force=True)
    db.load(dataset)
    cldf_ds = dataset.cldf_reader()
    cols = cldf_ds['FormTable'].tableSchema.columns
    cols.append(Column(name='custom'))
    cldf_ds.write_metadata()
    db.load(dataset)
    cols.pop()
    cols.append(Column(name='custom', datatype='integer'))
    cldf_ds.write_metadata()
    with pytest.raises(ValueError):
        db.load(dataset)
    cols.pop()
    cldf_ds.write_metadata()
    db.load(dataset)


def test_db_multiple_datasets(dataset, dataset_cldf, dataset_cldf_capitalisation, capsys, db):
    db.load(dataset)
    db.load(dataset_cldf, verbose=True)
    db.load(dataset_cldf_capitalisation, verbose=True)
    with db.connection() as conn:
        res = db.fetchall('select `id`, `name` from LanguageTable', conn=conn)
        assert len(res) == 6
        assert ('1', 'Lang CLDF') in [(r[0], r[1]) for r in res]
        res = db.fetchall('select `id`, `value` from FormTable', conn=conn)
        assert ('1', 'abc') in [(r[0], r[1]) for r in res]


def test_db_multiple_datasets_error(dataset, dataset_factory, db):
    import shutil
    from clldutils.jsonlib import load, dump

    assert not db.fname.exists()
    db.load(dataset)

    ds_dir = dataset.dir.parent / 'dbtest'
    shutil.copytree(str(dataset.dir), str(ds_dir))
    # Now modify the CLDF data:
    md = load(ds_dir / 'cldf' / 'cldf-metadata.json')
    for t in md['tables']:
        if t['url'] == 'parameters.csv':
            for col in t['tableSchema']['columns']:
                if col['name'] == 'Chinese':
                    col['name'] = 'chinese'
                    col['datatype'] = 'integer'
    dump(md, ds_dir / 'cldf' / 'cldf-metadata.json')

    ds2 = dataset_factory('dbtest.td')
    with pytest.raises(ValueError) as e:
        db.load(ds2)
        assert 'redefined' in str(e)



from argparse import Namespace

import pytest
import attr

from pylexibank.cldf import LexibankWriter
from pylexibank import Language, Dataset
from clldutils.jsonlib import load


def test_align_cognates(dataset, clts, mocker):
    with LexibankWriter(
        cldf_spec=dataset.cldf_specs(),
        dataset=dataset,
        args=Namespace(clts=mocker.Mock(api=clts))
    ) as ds:
        ds.add_language(ID='a-b')
        with pytest.raises(ValueError):
            ds.add_language(ID='a/b')
        lmap = ds.add_languages(lookup_factory='Name')
        assert isinstance(lmap, dict)
        assert lmap['abcd'] == '1'
        cids = ds.add_concepts()
        assert isinstance(cids, list)
        assert '1' in cids
        ds.align_cognates()

        with pytest.raises(ValueError):
            ds.add_form_with_segments()

        with pytest.raises(ValueError):
            ds.add_forms_from_value(Language_ID='l', Parameter_ID='p', Value='x', Segments=['x'])

        with pytest.raises(ValueError):
            ds.add_forms_from_value(Language_ID='l', Parameter_ID='p', Value='x', Form='x')

        with pytest.raises(ValueError):
            ds.add_form()

        with pytest.raises(ValueError):
            ds.add_form(
                Language_ID='l', Parameter_ID='p', Value='x', Form='x', Segments=['x'])

        with pytest.raises(ValueError):
            ds.add_concept(ID='1', Concepticon_ID='1', Concepticon_Gloss='xyz')

        ds.cldf['FormTable', 'Segments'].separator = '+'
        lex = ds.add_form_with_segments(
            Language_ID='l', Parameter_ID='p', Value='x', Form='x', Segments=['x', 'y'])
        assert lex['Segments'] == ['a', 'b']


def test_reqs(tmp_path, mocker, clts):
    class D(Dataset):
        dir = tmp_path
        id = 'x'

        def cmd_makecldf(self, args):
            args.writer.add_language(ID='l')
            args.writer.add_concept(ID='c')
            args.writer.add_form_with_segments(
                Language_ID='l', Parameter_ID='c', Value='x', Form='x', Segments=['x', 'y'])

    D()._cmd_makecldf(Namespace(
        log=mocker.Mock(), dev=False, verbose=False, clts=mocker.Mock(api=clts)))
    md =  load(tmp_path.joinpath('cldf', 'cldf-metadata.json'))
    reqs = {o['dc:title']: o for o in md['prov:wasGeneratedBy']}
    assert tmp_path.joinpath('cldf', reqs['lingpy-rcParams']['dc:relation']).exists()
    assert reqs['python-packages'] and reqs['python']


def test_custom_columns(tmp_path, clts, mocker):
    @attr.s
    class Variety(Language):
        """
        abcdefg
        """
        x = attr.ib(default=None, metadata={'separator': ';', 'dc:description': "-+-+-"})

    class D(Dataset):
        dir = tmp_path
        id = 'x'
        language_class = Variety
        writer_options = dict(keep_languages=True)

        def cmd_makecldf(self, args):
            args.writer.add_language(ID='l', x=['x', 'y'])
            args.writer.add_language(ID='l2', x=['x', 'y'])
            args.writer.add_concept(ID='c')
            args.writer.add_form_with_segments(
                Language_ID='l', Parameter_ID='c', Value='x', Form='x', Segments=['x', 'y'])

    D()._cmd_makecldf(Namespace(
        log=mocker.Mock(), dev=False, verbose=False, clts=mocker.Mock(api=clts)))
    langs = tmp_path.joinpath('cldf', 'languages.csv').read_text(encoding='utf8')
    assert 'x;y' in langs
    assert 'l2' in langs
    md = tmp_path.joinpath('cldf', 'cldf-metadata.json').read_text(encoding='utf8')
    assert '-+-+-' in md
    assert 'abcdefg' in md



"""
Create a profile for individual languages from an already prepared general profile.
"""
from collections import defaultdict

from cldfbench.cli_util import get_dataset
from clldutils.clilib import ParserError
from pycldf import Dataset
from csvw.dsv import reader, UnicodeWriter

from pylexibank import progressbar
from pylexibank.cli_util import add_dataset_spec


def register(parser):
    add_dataset_spec(parser)
    parser.add_argument(
        '-f', '--force',
        action='store_true',
        help='Overwrite existing profile',
        default=False)


def run(args):
    ds = get_dataset(args)
    wordlist = Dataset.from_metadata(ds.cldf_dir / "cldf-metadata.json")
    p = ds.etc_dir / "orthography.tsv"
    if not p.exists():
        raise ParserError("profile does not exist but is needed for creation")  # pragma: no cover
    if not ds.etc_dir.joinpath("orthography").exists():
        ds.etc_dir.joinpath("orthography").mkdir(parents=True, exist_ok=True)
    profile = {row["Grapheme"]: row["IPA"] for row in reader(p, dicts=True, delimiter='\t')}

    for language in progressbar(wordlist.objects("LanguageTable"), desc="creating profiles"):
        data = defaultdict(int)
        for form in language.forms:
            graphemes = form.data.get("Graphemes")
            if graphemes:
                for grapheme in (graphemes.split() if isinstance(graphemes, str) else graphemes):
                    data[grapheme, profile.get(grapheme, "?")] += 1
            else:
                raise ValueError("Grapheme information missing in CLDF data")
        new_path = ds.etc_dir / "orthography" / "{0}.tsv".format(language.id)
        if new_path.exists() and not args.force:  # pragma: no cover
            raise ParserError("Orthography profile exists, use --force to override")
        with UnicodeWriter(new_path, delimiter='\t') as writer:
            writer.writerow(["Grapheme", "IPA", "Frequency"])
            for (g, s), freq in sorted(data.items(), key=lambda x: x[1], reverse=True):
                writer.writerow([g, s, freq])
